(in-package "CCL")

;--------------- APPLICATION GLOBALS ------------------------------------

;------ ENVIRONMENT (GENERAL)

(defvar *file-path* nil)
(defvar *active-project-mode*)
(defvar *save-flag* nil)

;------ ENVIRONMENT (PREFERENCES)

(defvar *default-project-mode*)
(defvar *interface-serial-port*)
(defvar *default-window-size*)
(defvar *current-window-size*)
(defvar *size-borders-flag*)

;------ STRUCTURES

(defvar *save-item*)
(defvar *examples-menu*)
(defvar *window*)

;------ RESOURCES

(defvar *picts*)
(defvar *pict-buffers*)
(defvar *sounds*)

;------ CONSTANTS

(defconstant *version-string*     "2.5.4")
(defconstant *development-folder* (concatenate 'string "External (2GB):Programming Projects:LogoBlocks "
                                                       *version-string*
                                                       ":"))
(defconstant *dialog-color*       *tool-back-color*)
(defconstant *number-of-picts*    76)
(defconstant *number-of-sounds*    5)

;--------------- APPLICATION CLASSES/SSMs ------------------------------------

(defclass logoblocks-application (application)
  ())

(defmethod application-name ((application logoblocks-application))
  (declare (ignore application))
  (concatenate 'string "LogoBlocks " *version-string* " (- Libs)"))

(defmethod application-file-creator ((application logoblocks-application))
  (declare (ignore application))
  :|LgBl|)

(defmethod application-sizes ((application logoblocks-application))
  (declare (ignore application))
  (list (* 12000 1024)
        (* 20000 1024)))

(defmethod application-resource-file ((application logoblocks-application))
  (declare (ignore application))
  (concatenate 'string *development-folder* "complete.rsrc"))

(defmethod application-about-dialog ((application logoblocks-application))
  (declare (ignore application))
  (about-dialog))

(defmethod open-application-document ((application logoblocks-application) filename &optional startup)
  (declare (ignore application startup))
  (open-file *window* filename))

;--------------- APPLICATION CONSTRUCTORS ------------------------------------

(defun build-application ()
  (modify-lisp-environment)
  (load-source-files)
  (create-logoblocks-menubar)
  (setf *lisp-startup-functions* (list #'(lambda () (load-resources) (setup-environment))))
  (setf *application*            (make-instance 'logoblocks-application))
  (let ((path (choose-new-file-dialog :prompt    "Save As"
                                      :directory (concatenate 'string *development-folder*
                                                                      (application-name *application*)))))
    (save-application path :creator           (application-file-creator   *application*)
                           :size              (reverse (application-sizes *application*))
                           :resources         (get-app-resources (application-resource-file *application*)
                                                                 (application-file-creator  *application*))
                           :toplevel-function nil
                           :init-file         nil
                           :error-handler     :listener
                           :excise-compiler   t
                           :clear-CLOS-caches t)))

;---------------

(defun modify-lisp-environment ()
  (format t "~%~%Loading Patch: y-or-n-dialog.")       (load "ccl:Patches 3.3;y-or-n-dialog-patch.lisp"))

(defun load-source-files ()
  (format t "~%~%Compiling Unit: Toolbox")             (load (concatenate 'string *development-folder* "toolbox.lisp"))
  (format t   "~%Compiling Unit: Menus/Windows/Views") (load (concatenate 'string *development-folder* "menus/windows/views.lisp"))
  (format t   "~%Compiling Unit: Blocks")              (load (concatenate 'string *development-folder* "blocks.lisp"))
  (format t   "~%Compiling Unit: Graphics")            (load (concatenate 'string *development-folder* "graphics.lisp"))
  (format t   "~%Compiling Unit: Dialogs")             (load (concatenate 'string *development-folder* "dialogs.lisp"))
  (format t   "~%Compiling Unit: Files")               (load (concatenate 'string *development-folder* "files.lisp"))
  (format t   "~%Compiling Unit: Hardware")            (load (concatenate 'string *development-folder* "hardware.lisp"))
  (format t   "~%Compiling Unit: Compiler")            (load (concatenate 'string *development-folder* "compiler.lisp")))

(defun load-resources ()
  (load-picts)
  (alloc/init-pict-buffers)
  (dispose-picts)
  (load-sounds))

(defun setup-environment ()
  (read-preferences)
  (rebuild-examples-menu)
  (create-logoblocks-window)
  (open-interface-serial-stream))

;--------------- APPLICATION INITIATORS ------------------------------------

(build-application)