;--------------- BLOCK CLASSES ------------------------------------

(defclass logoblocks-block ()
  ((boundrect                :initarg :boundrect :accessor boundrect)
   (boundrgn                 :initarg :boundrgn  :accessor boundrgn)
   (species                  :initarg :species   :reader   species)
   (family                   :initarg :family    :reader   family)
   (visible?   :initform t                       :accessor visible?)
   (fulltext?                :initarg :fulltext? :accessor fulltext?)
   (selected?                :initarg :selected? :accessor selected?)
   (docks                    :initarg :docks     :reader   docks)))

(defclass separator-block (logoblocks-block)
  ())

;---------------

(defclass on/off-block (logoblocks-block)
  ((command    :initform 1                       :accessor command)
   (motorA?    :initform t                       :accessor motorA?)
   (motorB?    :initform t                       :accessor motorB?)))

(defclass direction-block (logoblocks-block)
  ((command    :initform 1                       :accessor command)
   (motorA?    :initform t                       :accessor motorA?)
   (motorB?    :initform t                       :accessor motorB?)))

(defclass power-block (logoblocks-block)
  ((command    :initform 8                       :accessor command)
   (motorA?    :initform t                       :accessor motorA?)
   (motorB?    :initform t                       :accessor motorB?)))

(defclass onfor-block (logoblocks-block)
  ((motorA?    :initform t                       :accessor motorA?)
   (motorB?    :initform t                       :accessor motorB?)))

(defclass beep-block (logoblocks-block)
  ())

(defclass note-block (logoblocks-block)
  ((pitch      :initform 3                       :accessor pitch)
   (modifier   :initform 2                       :accessor modifier)
   (octave     :initform 2                       :accessor octave)))

(defclass wait-block (logoblocks-block)
  ())

(defclass zerotimer-block (logoblocks-block)
  ())

(defclass timer-block (logoblocks-block)
  ())

(defclass dummy-block (logoblocks-block)
  ())

;---------------

(defclass when-block (logoblocks-block)
  ())

(defclass whenoff-block (logoblocks-block)
  ())

(defclass if-block (logoblocks-block)
  ())

(defclass ifelse-block (logoblocks-block)
  ())

(defclass repeat-block (logoblocks-block)
  ())

(defclass loop-block (logoblocks-block)
  ())

(defclass waituntil-block (logoblocks-block)
  ())

;---------------

(defclass switch-block (logoblocks-block)
  ((switch     :initform 1                       :accessor switch)))

(defclass sensor-block (logoblocks-block)
  ((sensor     :initform 1                       :accessor sensor)))

(defclass clearir-block (logoblocks-block)
  ())

(defclass newir?-block (logoblocks-block)
  ())

(defclass ir-block (logoblocks-block)
  ())

(defclass sendir-block (logoblocks-block)
  ())

(defclass display-block (logoblocks-block)
  ((color      :initform 4                       :accessor color)))

(defclass number-block (logoblocks-block)
  ((genmethod  :initform 1                       :accessor genmethod)
   (value      :initform 0                       :accessor value)
   (lowerbound :initform 0                       :accessor lowerbound)
   (upperbound :initform 255                     :accessor upperbound)))

(defclass boolean-block (logoblocks-block)
  ((comparator :initform 3                       :accessor comparator)))

(defclass and-block (logoblocks-block)
  ())

(defclass or-block (logoblocks-block)
  ())

(defclass cleardata-block (logoblocks-block)
  ())

(defclass record-block (logoblocks-block)
  ())

(defclass resetdp-block (logoblocks-block)
  ())

;---------------

(defclass defproc-block (logoblocks-block)
  ((procnum                                      :accessor procnum)
   (proctitle                                    :accessor proctitle)
   (mempos     :initform nil                     :accessor mempos)))

(defclass stop-block (logoblocks-block)
  ())

(defclass callproc-block (logoblocks-block)
  ((procnum    :initform 0                       :accessor procnum)
   (proctitle  :initform "None"                  :accessor proctitle)))

;--------------- BLOCK CREATION OPS ------------------------------------

(defmethod create-block ((view working-view) species fulltext? selected? position &optional (extra-stuff nil))
  (let* ((mask-buffer (nth (+ (* (block-family species fulltext?) 3) 2) *pict-buffers*))
         (block       (make-instance (block-class-symbol species) :boundrect (duplicate-rect (access-gworld-boundrect mask-buffer))
                                                                  :boundrgn                  (create-gworld-boundrgn  mask-buffer)
                                                                  :species   species
                                                                  :family    (block-family species fulltext?)
                                                                  :fulltext? fulltext?
                                                                  :selected? selected?
                                                                  :docks     (make-array 5 :initial-contents (list 'none 'none 'none 'none 'none)))))
    (move-blocks-to (list block) position)
    (let ((docks (block-docks (family block))))
      (loop for x in (first-n-integers (length docks)) do
            (setf (aref (docks block) x) (nth x docks))))
    (if extra-stuff
      (if (equal species 'number)
        (setf (value block) extra-stuff)
        (assign-procedure-block-fields view block extra-stuff)))
    (setf (masterlist view) (append (masterlist view) (list block)))
    block))

;---------------

(defmethod assign-procedure-block-fields ((view palette-view) block procdata)
  (declare (ignore view))
  (setf (procnum   block) (car  procdata))
  (setf (proctitle block) (cadr procdata)))

(defmethod assign-procedure-block-fields ((view wrkarea-view) block procdata)
  (cond ((equal (species block) 'defproc)  (if (> (car procdata) 0)
                                             (progn (setf (procnum   block) (car  procdata))
                                                    (setf (proctitle block) (cadr procdata)))
                                             (progn (setf (procnum   block) (+ (highest-procnum-deployed view) 1))
                                                    (setf (proctitle block) (concatenate 'string "?? " (format nil "~S" (- (procnum block) 2)))))))
        ((equal (species block) 'callproc) (setf (procnum   block) (car procdata))
                                           (setf (proctitle block) (if (= (procnum block) 0) "None" (cadr procdata))))))

;--------------- BLOCK SELECTION OPS ------------------------------------

(defmethod select-block ((view wrkarea-view) block &optional (grab-numbers? nil))
  (if (and grab-numbers?
           (or (equal (family block)  1)
               (equal (family block)  7)
               (equal (family block) 12)
               (equal (family block) 15)))
    (let ((block1 (scan-dock view block 1))
          (block2 (if (equal (family block) 12)
                    (scan-dock view block 2))))
      (if block1 (progn (setf (selected? block1) t)
                        (draw-buffer3 view (list block1) 'draw)))
      (if block2 (progn (setf (selected? block2) t)
                        (draw-buffer3 view (list block2) 'draw)))))
  (setf (selected? block) t)
  (draw-buffer3 view (list block) 'draw)
  (view-draw-contents view))

(defmethod select-all-blocks ((view wrkarea-view))
  (loop for block in (masterlist view) do
        (setf (selected? block) t))
  (draw-buffer3       view (masterlist view) 'draw)
  (view-draw-contents view))

(defmethod deselect-all-blocks ((view wrkarea-view))
  (loop for block in (masterlist view) do
        (setf (selected? block) nil))
  (draw-buffer3       view (masterlist view) 'draw)
  (view-draw-contents view))

(defmethod invert-block-selections ((view wrkarea-view) blocklist)
  (loop for block in blocklist do
        (if (selected? block)
          (setf (selected? block) nil)
          (setf (selected? block) t)))
  (draw-buffer3       view blocklist 'draw)
  (view-draw-contents view))

(defmethod get-selected-blocks ((view wrkarea-view))
  (loop for block in (masterlist view)
        if (selected? block) collect block))

;--------------- BLOCK DESTRUCTION OPS ------------------------------------

(defmethod delete-all-blocks ((view working-view))
  (draw-buffer3       view (masterlist view) 'erase)
  (dispose-blocks     view (masterlist view))
  (view-draw-contents view))

(defmethod rebuild-all-blocks ((view working-view))
  (reset-view         view nil)
  (draw-buffer3       view (masterlist view) 'draw)
  (view-draw-contents view))

(defmethod redo-all-blocks ((view palette-view))
  (draw-buffer3       view (masterlist view) 'erase)
  (dispose-blocks     view (masterlist view))
  (reset-view         view nil)
  (draw-buffer3       view (masterlist view) 'draw)
  (view-draw-contents view))

(defmethod delete-selected-blocks ((view wrkarea-view))
  (let ((blocklist (set-difference (get-selected-blocks view)
                                   (list (car (masterlist view)))))
        (flag1     nil)
        (flag2     nil))
    (loop for block1 in blocklist do
          (cond ((equal (species block1) 'when)
                 (setf flag1 t))
                ((equal (species block1) 'defproc)
                 (setf flag2 t)
                 (loop for block2 in (masterlist view) do
                       (if (equal (species block2) 'callproc)
                         (if (= (procnum block2) (procnum block1))
                           (progn (setf (procnum   block2) 0)
                                  (setf (proctitle block2) "None")
                                  (draw-buffer3 view (list block2) 'draw))))
                       (if (or (equal (species block2) 'defproc)
                               (equal (species block2) 'callproc))
                         (if (> (procnum block2) (procnum block1) 2)
                           (progn (setf (procnum block2) (- (procnum block2) 1))
                                  (if (equal     (proctitle block2) (concatenate 'string "?? " (format nil "~S" (- (procnum block2) 1))))
                                    (progn (setf (proctitle block2) (concatenate 'string "?? " (format nil "~S" (- (procnum block2) 2))))
                                           (draw-buffer3 view (list block2) 'draw))))))))))
    (draw-buffer3       view blocklist 'erase)
    (dispose-blocks     view blocklist)
    (view-draw-contents view)
    (if (or (and flag1 (= (switch (buttons (view-container view))) 1))
            (and flag2 (= (switch (buttons (view-container view))) 3)))
      (redo-all-blocks (palette (view-container view))))))

;---------------

(defmethod dispose-blocks ((view working-view) blocklist)
  (setf (masterlist view) (reverse (set-difference (masterlist view) blocklist)))
  (loop for block in blocklist do
        (dispose-record (boundrect block))
        (dispose-region (boundrgn  block))))

;--------------- BLOCK REPOSITION OPS ------------------------------------

(defmethod scrollset-blocks ((view palette-view))
  (let ((offset (- 10 (top (boundrect (nth (scroll-bar-setting (scrollbar (view-container view))) (masterlist view)))))))
    (offset-blocks (masterlist view) (make-point 0 offset))
    (set-block-visibilities view)))

(defmethod drag-blocks ((view palette-view) blocklist from)
  (if (execute-drag  (wrkarea (view-container view)) blocklist from)
    (if (snap-blocks (wrkarea (view-container view)) blocklist #@(0 0))
      (progn (draw-buffer3       (wrkarea (view-container view)) blocklist 'draw)
             (view-draw-contents (wrkarea (view-container view)))
             (if (or (equal (species (car blocklist)) 'when)
                     (equal (species (car blocklist)) 'defproc))
               (redo-all-blocks view)))
      (progn (dispose-blocks     (wrkarea (view-container view)) blocklist)
             (view-draw-contents (wrkarea (view-container view)))
             (if (or      (equal (species (car blocklist)) 'when)
                     (and (equal (species (car blocklist)) 'defproc)
                          (=     (procnum (car blocklist)) 2)))
               (progn (draw-buffer3       view (list (car (masterlist view))) 'draw)
                      (view-draw-contents view)))))
    (progn (dispose-blocks (wrkarea (view-container view)) blocklist)
           (if (or      (equal (species (car blocklist)) 'when)
                   (and (equal (species (car blocklist)) 'defproc)
                        (=     (procnum (car blocklist)) 2)))
             (draw-buffer3 view (list (car (masterlist view))) 'draw))
           (view-draw-contents (wrkarea   (view-container view)))
           (view-draw-contents (palette   (view-container view)))
           (view-draw-contents (scrollbar (view-container view))))))

(defmethod drag-blocks ((view wrkarea-view) blocklist from)
  (draw-buffer3       view blocklist 'erase)
  (snap-blocks        view blocklist (execute-drag view blocklist from))
  (draw-buffer3       view blocklist 'draw)
  (view-draw-contents view))

;--------------- BLOCK LINK OPS ------------------------------------

(defmethod snap-blocks ((view wrkarea-view) blocklist restorative)
  (let* ((staticset   (set-difference (masterlist view) blocklist))
         (forcedjump? (overlap? blocklist #@(0 0) staticset #@(0 0)))
         (bestrange   32)
         (localoffset nil))
    (loop for block1 in blocklist do
          (loop for block2 in staticset do
                (loop for x in (first-n-integers 5) do
                      (loop for y in (first-n-integers 5) do
                            (if (potential-link? view block1 x block2 y)
                              (let ((results (link-range-and-offset block1 x block2 y)))
                                (if (< (car results) bestrange)
                                  (unless (or (overlap? blocklist (cadr results) staticset #@(0 0))
                                              (out-of-bounds? view blocklist (cadr results)))
                                    (setf bestrange   (car  results))
                                    (setf localoffset (cadr results))))))))))
    (if localoffset (play-sound-sync 'snap)
                    (if forcedjump? (play-sound-sync 'whisk)))
    (let ((jump     (if localoffset localoffset
                                    (if forcedjump? restorative
                                                    #@(0 0))))
          (success  (not (and (not localoffset) forcedjump?))))
      (offset-blocks blocklist jump)
      success)))

;---------------

(defun overlap? (blocklist1 offset1 blocklist2 offset2)
  (let ((group1 (new-region))
        (group2 (new-region)))
    (loop for block in blocklist1 do
          (union-regions group1 (boundrgn block) group1))
    (loop for block in blocklist2 do
          (union-regions group2 (boundrgn block) group2))
    (offset-region group1 offset1)
    (offset-region group2 offset2)
    (let ((result (not (empty-region? (intersect-regions group1 group2)))))
      (dispose-region group1)
      (dispose-region group2)
      result)))

(defmethod out-of-bounds? ((view wrkarea-view) blocklist offset)
  (let ((blockgroup (new-region))
        (result     nil))
    (loop for block in blocklist do
          (union-regions blockgroup (boundrgn block) blockgroup))
    (offset-region blockgroup offset)
    (let ((groupbounds (access-region-boundrect blockgroup)))
      (#_InsetRect groupbounds -1 -1)
      (if (or (< (left   groupbounds) 0)
              (> (right  groupbounds) (point-h (view-size view)))
              (< (top    groupbounds) 0)
              (> (bottom groupbounds) (point-v (view-size view))))
        (setf result t)))
    (dispose-region blockgroup)
    result))

(defmethod potential-link? ((view wrkarea-view) block1 x block2 y)
  (let ((dock1 (aref (docks block1) x))
        (dock2 (aref (docks block2) y))
        (result nil))
    (if (or (and (equal dock1 'up)        (equal dock2 'down))
            (and (equal dock1 'up)        (equal dock2 'sequence1))
            (and (equal dock1 'up)        (equal dock2 'sequence2))
            (and (equal dock1 'down)      (equal dock2 'up))
            (and (equal dock1 'sequence1) (equal dock2 'up))
            (and (equal dock1 'sequence2) (equal dock2 'up))
            (and (equal dock1 'number)    (equal dock2 'number1))
            (and (equal dock1 'number)    (equal dock2 'number2))
            (and (equal dock1 'number1)   (equal dock2 'number))
            (and (equal dock1 'number2)   (equal dock2 'number))
            (and (equal dock1 'bool-L-M)  (equal dock2 'bool-R-F))
            (and (equal dock1 'bool-R-M)  (equal dock2 'bool-L-F))
            (and (equal dock1 'bool-L-F)  (equal dock2 'bool-R-M))
            (and (equal dock1 'bool-R-F)  (equal dock2 'bool-L-M)))
      (if (and (not (scan-dock view block1 x))
               (not (scan-dock view block2 y)))
        (setf result t)))
    result))

(defun link-range-and-offset (block1 x block2 y)
  (let ((position1 (add-points (topleft (boundrect block1))
                               (dock-offset (family block1) (aref (docks block1) x))))
        (position2 (add-points (topleft (boundrect block2))
                               (dock-offset (family block2) (aref (docks block2) y)))))
    (list (distance        position1 position2)
          (subtract-points position2 position1))))

;--------------- MISCELLANEOUS BLOCK OPS ------------------------------------

(defun block-class-symbol (species)
  (cond ((equal species 'separator) 'separator-block)

        ((equal species 'on/off)    'on/off-block)
        ((equal species 'direction) 'direction-block)
        ((equal species 'power)     'power-block)
        ((equal species 'onfor)     'onfor-block)
        ((equal species 'beep)      'beep-block)
        ((equal species 'note)      'note-block)
        ((equal species 'wait)      'wait-block)
        ((equal species 'zerotimer) 'zerotimer-block)
        ((equal species 'timer)     'timer-block)
        ((equal species 'dummy)     'dummy-block)

        ((equal species 'when)      'when-block)
        ((equal species 'whenoff)   'whenoff-block)
        ((equal species 'if)        'if-block)
        ((equal species 'ifelse)    'ifelse-block)
        ((equal species 'repeat)    'repeat-block)
        ((equal species 'loop)      'loop-block)
        ((equal species 'waituntil) 'waituntil-block)

        ((equal species 'switch)    'switch-block)
        ((equal species 'sensor)    'sensor-block)
        ((equal species 'clearir)   'clearir-block)
        ((equal species 'newir?)    'newir?-block)
        ((equal species 'ir)        'ir-block)
        ((equal species 'sendir)    'sendir-block)
        ((equal species 'display)   'display-block)
        ((equal species 'number)    'number-block)
        ((equal species 'boolean)   'boolean-block)
        ((equal species 'and)       'and-block)
        ((equal species 'or)        'or-block)
        ((equal species 'cleardata) 'cleardata-block)
        ((equal species 'record)    'record-block)
        ((equal species 'resetdp)   'resetdp-block)

        ((equal species 'defproc)   'defproc-block)
        ((equal species 'stop)      'stop-block)
        ((equal species 'callproc)  'callproc-block)))

;  0 Action Small
;  1 Action Small+N
;  2 Action Deeper
;
;  3 Control When
;  4 Control Deeper
;  5 Control If
;  6 Control IfElse
;  7 Control Repeat
;  8 Control Loop
;  9 Control WaitUntil
;
; 10 Sensor Bool
; 11 Sensor WideBoolComp
; 12 Sensor WideBool
; 13 Sensor BoolOp
; 14 Sensor Small
; 15 Sensor Small+N
; 16 Sensor Deeper
;
; 17 Procedure DefProc
; 18 Procedure Stop
; 19 Procedure CallProc
;
; 20 Number Blue
; 21 Number Green
; 22 Number Red
;
; 23 Separator

(defun block-family (species &optional (wrkarea? t))
  (cond ((equal species 'separator) 23)

        ((equal species 'on/off)     0)
        ((equal species 'direction)  0)
        ((equal species 'power)      0)
        ((equal species 'onfor)      1)
        ((equal species 'beep)       0)
        ((equal species 'note)       1)
        ((equal species 'wait)       1)
        ((equal species 'zerotimer)  2)
        ((equal species 'timer)     21)
        ((equal species 'dummy)      0)

        ((equal species 'when)       3)
        ((equal species 'whenoff)    4)
        ((equal species 'if)         5)
        ((equal species 'ifelse)     (if wrkarea? 6 5))
        ((equal species 'repeat)     7)
        ((equal species 'loop)       8)
        ((equal species 'waituntil)  9)

        ((equal species 'switch)    10)
        ((equal species 'sensor)    22)
        ((equal species 'clearir)   16)
        ((equal species 'newir?)    10)
        ((equal species 'ir)        22)
        ((equal species 'sendir)    15)
        ((equal species 'display)   15)
        ((equal species 'number)    20)
        ((equal species 'boolean)   (if wrkarea? 12 11))
        ((equal species 'and)       13)
        ((equal species 'or)        13)
        ((equal species 'cleardata) 16)
        ((equal species 'record)    15)
        ((equal species 'resetdp)   14)

        ((equal species 'defproc)   17)
        ((equal species 'stop)      18)
        ((equal species 'callproc)  19)))

(defun block-h-offset (switch family)
  (if (or (= family 11)
          (= family 23)
          (= switch 1))
    10 30))

(defun block-docks (family)
  (cond ((or (= family  0)
             (= family  2)
             (= family  4)
             (= family 14)
             (= family 16)
             (= family 19)) (list 'up 'down))
        ((or (= family  1)
             (= family 15)) (list 'up 'number1 'down))

        (    (= family  3)  (list     'bool-R-F                       'down))
        (    (= family  5)  (list 'up 'bool-R-F 'sequence1            'down))
        (    (= family  6)  (list 'up 'bool-R-F 'sequence1 'sequence2 'down))
        (    (= family  7)  (list 'up 'number1  'sequence1            'down))
        (    (= family  8)  (list 'up           'sequence1            'down))
        (    (= family  9)  (list 'up 'bool-R-F                       'down))

        (    (= family 10)  (list 'bool-L-M                   'bool-R-M))
        (    (= family 12)  (list 'bool-L-M 'number1 'number2 'bool-R-M))
        (    (= family 13)  (list 'bool-L-F 'bool-R-F))

        (    (= family 17)  (list 'down))
        (    (= family 18)  (list 'up  ))

        ((or (= family 20)
             (= family 21)
             (= family 22)) (list 'number))

        (    (= family 23)  (list))))

(defun dock-offset (family docktype)
  (cond ((equal docktype 'up)                                  #@( 18  0))
        ((equal docktype 'down)      (cond ((or (= family  0)
                                                (= family  1)
                                                (= family  3)
                                                (= family  9)
                                                (= family 14)
                                                (= family 15)
                                                (= family 17)
                                                (= family 19)) #@( 18 26))
                                           ((or (= family  2)
                                                (= family  4)
                                                (= family 16)) #@( 18 33))
                                           (    (= family  5)  #@( 18 62))
                                           (    (= family  6)  #@( 18 76))
                                           (    (= family  7)  #@( 18 50))
                                           (    (= family  8)  #@( 18 28))))
        ((equal docktype 'number)                              #@(  0  0))
        ((equal docktype 'number1)   (cond ((or (= family  1)
                                                (= family 15)) #@( 50  6))
                                           (    (= family  7)  #@( 50  4))
                                           (    (= family 12)  #@( 14  6))))
        ((equal docktype 'number2)                             #@( 86  6))
        ((equal docktype 'bool-L-M)                            #@(  8 12))
        ((equal docktype 'bool-R-M)  (cond (    (= family 10)  #@( 40 12))
                                           (    (= family 12)  #@(144 12))))
        ((equal docktype 'bool-L-F)                            #@(  8 12))
        ((equal docktype 'bool-R-F)  (cond ((or (= family  3)
                                                (= family  9)) #@( 62 12))
                                           ((or (= family  5)
                                                (= family  6)) #@( 62 16))
                                           (    (= family 13)  #@( 40 12))))
        ((equal docktype 'sequence1) (cond (    (= family  5)  #@( 72 36))
                                           (    (= family  6)  #@( 72 50))
                                           (    (= family  7)  #@( 72 24))
                                           (    (= family  8)  #@( 72  2))))
        ((equal docktype 'sequence2)                           #@(126 50))
        ((equal docktype 'none)                                #@(  0  0))))

(defmethod scan-dock ((view wrkarea-view) block1 x)
  (let ((result nil))
    (loop for block2 in (set-difference (masterlist view) (list block1)) do
          (loop for y in (first-n-integers 5) do
                (if (= (add-points (topleft (boundrect block1))
                                   (dock-offset (family block1) (aref (docks block1) x)))
                       (add-points (topleft (boundrect block2))
                                   (dock-offset (family block2) (aref (docks block2) y))))
                  (setf result block2))))
    result))

;---------------

(defmethod set-block-visibilities ((view palette-view))
  (loop for block in (masterlist view) do
        (if (or (< (top    (boundrect block)) 10)
                (> (bottom (boundrect block)) (- (point-v (view-size view)) 12)))
          (setf (visible? block) nil)
          (setf (visible? block) t))))

(defmethod count-lower-invisibles ((view palette-view))
  (let ((x        0)
        (proceed? t))
    (loop while proceed? do
          (if (visible? (nth x (reverse (masterlist view))))
            (setf proceed? nil)
            (setf x (+ x 1))))
    x))

;---------------

(defmethod acquire-target ((view working-view) where)
  (loop for block in (masterlist view)
        if (and (visible? block)
                (point-in-region? where (boundrgn block))
                (not (equal (species block) 'separator)))
        return block))

(defmethod blocks-in-rect ((view wrkarea-view) rect)
  (loop for block in (masterlist view)
        if (rect-in-region? rect (boundrgn block))
        collect block))

;---------------

(defmethod index-defproc-block ((view wrkarea-view) procnum)
  (let ((result nil))
    (loop for block in (masterlist view) do
          (if (equal (species block) 'defproc)
            (if (= (procnum block) procnum)
              (setf result block))))
    result))

(defmethod highest-procnum-deployed ((view wrkarea-view))
  (let ((highscore 2))
    (loop for block in (masterlist view) do
          (if (equal (species block) 'defproc)
            (if (> (procnum block) highscore)
              (setf highscore (procnum block)))))
    highscore))

(defmethod valid-proctitle? ((view wrkarea-view) proctitle)
  (let ((result t))
    (if (or (equal proctitle "P2")
            (equal proctitle ""))
      (setf result nil)
      (loop for block in (masterlist view) do
            (if (equal (species block) 'defproc)
              (if (equal (proctitle block) proctitle)
                (setf result nil)))))
    result))

(defmethod cluster-root ((view wrkarea-view) block)
  (let ((block0 (scan-dock view block 0)))
    (if (or (equal (species block) 'defproc)
            (not block0))
      block
      (cluster-root view block0))))

;---------------

(defmethod index-when-block ((view wrkarea-view))
  (let ((result nil))
    (loop for block in (masterlist view) do
          (if (equal (species block) 'when)
            (setf result block)))
    result))

;---------------

(defmethod LtG ((thing cons) view)
  (offset-blocks thing (view-position view)))

(defmethod GtL ((thing cons) view)
  (offset-blocks thing (inverse (view-position view))))

(defun offset-blocks (blocklist offset)
  (loop for block in blocklist do
        (offset-rect   (boundrect block) offset)
        (offset-region (boundrgn  block) offset)))

(defun move-blocks-to (blocklist position)
  (loop for block in blocklist do
        (move-rect-to   (boundrect block) position)
        (move-region-to (boundrgn  block) position)))