;--------------- HIGH-LEVEL OPS ------------------------------------

(defmethod download-p1/p2 ((view wrkarea-view))
  (let ((flag1 (all-key-docks-filled?    view))
        (flag2 (no-recursive-proc-calls? view)))
    (if (and flag1 flag2)
      (if (sufficient-partition1-memory? view)
        (download-message (download-all-proc-clusters view))
        (memory-error))
      (initial-pass-error flag1 flag2))))

;---------------

(defun initial-pass-error (flag1 flag2)
  (beep)
  (alert-dialog "Error" (concatenate 'string (if (not flag1) "All NUMBER, BOOLEAN and SEQUENCE docks
must be filled.

")
                                             (if (not flag2) "Recursive procedure calls are not supported
by this software.")) 400 134))

(defun memory-error ()
  (beep)
  (alert-dialog "Error" (cond ((= *active-project-mode* 1) "Too many blocks for a red dot cricket." )
                              ((= *active-project-mode* 2) "Too many blocks for a blue dot cricket.")
                              ((= *active-project-mode* 3) "Too many blocks for a LEGO RCX Brick."  ))
                        335 62))

(defun download-message (result)
  (cond ((= result -4) (play-sound-sync 'radar)
                       (alert-dialog "Error" (cond ((= *active-project-mode* 2) "Either there's no interface cricket, or it's not turned on."))
                                             310 80))
        ((= result -3) (play-sound-sync 'radar)
                       (alert-dialog "Error" (cond ((= *active-project-mode* 1) "No interface cricket detected.")
                                                   ((= *active-project-mode* 3) "No interface brick detected."  ))
                                             290 62))
        ((= result -2) (play-sound-sync 'radar)
                       (alert-dialog "Error" (cond ((= *active-project-mode* 1) "The interface cricket is not turned on.")
                                                   ((= *active-project-mode* 3) "The interface brick is not turned on."  ))
                                             335 62))
        ((= result -1) (play-sound-sync 'radar)
                       (alert-dialog "Error" (cond ((= *active-project-mode* 1) "No red dot crickets in sight." )
                                                   ((= *active-project-mode* 2) "No blue dot crickets in sight." )
                                                   ((= *active-project-mode* 3) "No LEGO RCX Bricks in sight."  ))
                                             290 62))
        ((= result  0) (play-sound-sync 'chirp)
                       (alert-dialog "Success" "Download successful." 265 62))))

;--------------- PHASE 1 (SOFTWARE) CHECKS ------------------------------------

(defmethod all-key-docks-filled? ((view wrkarea-view))
  (let ((proceed? t))
    (loop for block in (masterlist view) do
          (if (or (equal (species block) 'when)
                  (equal (species block) 'defproc))
            (if (not (check-from-block view block))
              (setf proceed? nil))))
    proceed?))

(defmethod no-recursive-proc-calls? ((view wrkarea-view))
  (let ((proceed? t))
    (loop for block in (masterlist view) do
          (if (equal (species block) 'callproc)
            (let ((root (cluster-root view block)))
              (if (and (equal (species root) 'defproc)
                       (> (procnum block) 0))
                (let ((trace (index-defproc-block view (procnum block))))
                  (if (eq trace root)
                    (setf proceed? nil)
                    (if (not (scan-from-block view trace (list (procnum trace)) root))
                      (setf proceed? nil))))))))
    proceed?))

;---------------

(defmethod check-from-block ((view wrkarea-view) block)
  (let ((block0 (scan-dock view block 0))
        (block1 (scan-dock view block 1))
        (block2 (scan-dock view block 2))
        (block3 (scan-dock view block 3))
        (block4 (scan-dock view block 4))
        (valid? t))
    (cond ((or (= (family block)  0)
               (= (family block)  2)
               (= (family block)  4)
               (= (family block) 10)
               (= (family block) 14)
               (= (family block) 16)
               (= (family block) 19)) (if block1 (if (not (check-from-block view block1)) (setf valid? nil))))
          ((or (= (family block)  1)
               (= (family block)  8)
               (= (family block)  9)
               (= (family block) 15)) (if block1 (if (not (check-from-block view block1)) (setf valid? nil)) (setf valid? nil))
                                      (if block2 (if (not (check-from-block view block2)) (setf valid? nil))))
          (    (= (family block)  3)  (if block0 (if (not (check-from-block view block0)) (setf valid? nil)) (setf valid? nil))
                                      (if block1 (if (not (check-from-block view block1)) (setf valid? nil))))
          ((or (= (family block)  5)
               (= (family block)  7)
               (= (family block) 12)) (if block1 (if (not (check-from-block view block1)) (setf valid? nil)) (setf valid? nil))
                                      (if block2 (if (not (check-from-block view block2)) (setf valid? nil)) (setf valid? nil))
                                      (if block3 (if (not (check-from-block view block3)) (setf valid? nil))))
          (    (= (family block)  6)  (if block1 (if (not (check-from-block view block1)) (setf valid? nil)) (setf valid? nil))
                                      (if block2 (if (not (check-from-block view block2)) (setf valid? nil)) (setf valid? nil))
                                      (if block3 (if (not (check-from-block view block3)) (setf valid? nil)) (setf valid? nil))
                                      (if block4 (if (not (check-from-block view block4)) (setf valid? nil))))
          (    (= (family block) 13)  (if block1 (if (not (check-from-block view block1)) (setf valid? nil)) (setf valid? nil)))
          ((or (= (family block) 18)
               (= (family block) 20)
               (= (family block) 21)
               (= (family block) 22)) ())
          (    (= (family block) 17)  (if block0 (if (not (check-from-block view block0)) (setf valid? nil)))))
    valid?))

(defmethod scan-from-block ((view wrkarea-view) block locationlog root)
  (let ((block0 (scan-dock view block 0))
        (block1 (scan-dock view block 1))
        (block2 (scan-dock view block 2))
        (block3 (scan-dock view block 3))
        (block4 (scan-dock view block 4))
        (result t))
    (cond ((or (= (family block)  0)
               (= (family block)  2)
               (= (family block)  3)
               (= (family block)  4)
               (= (family block) 14)
               (= (family block) 16)) (if block1 (if (not (scan-from-block view block1 locationlog root)) (setf result nil))))

          ((or (= (family block)  1)
               (= (family block)  9)
               (= (family block) 15)) (if block2 (if (not (scan-from-block view block2 locationlog root)) (setf result nil))))

          ((or (= (family block)  5)
               (= (family block)  7)) (if block2 (if (not (scan-from-block view block2 locationlog root)) (setf result nil)))
                                      (if block3 (if (not (scan-from-block view block3 locationlog root)) (setf result nil))))
          (    (= (family block)  6)  (if block2 (if (not (scan-from-block view block2 locationlog root)) (setf result nil)))
                                      (if block3 (if (not (scan-from-block view block3 locationlog root)) (setf result nil)))
                                      (if block4 (if (not (scan-from-block view block4 locationlog root)) (setf result nil))))

          (    (= (family block)  8)  (if block1 (if (not (scan-from-block view block1 locationlog root)) (setf result nil)))
                                      (if block2 (if (not (scan-from-block view block2 locationlog root)) (setf result nil))))

          (    (= (family block) 17)  (if block0 (if (not (scan-from-block view block0 locationlog root)) (setf result nil))))
          (    (= (family block) 18)  ())
          (    (= (family block) 19)  (if (> (procnum block) 0)
                                        (let ((trace (index-defproc-block view (procnum block))))
                                          (if (eq trace root)
                                            (setf result nil)
                                            (unless (match? (procnum trace) locationlog)
                                              (if (not (scan-from-block view trace (append locationlog (list (procnum trace))) root))
                                                (setf result nil))))))
                                      (if block1 (if (not (scan-from-block view block1 locationlog root)) (setf result nil)))))
    result))

;--------------- PHASE 2 (HARDWARE) CHECKS ------------------------------------

; Red Dot Cricket EEPROM Memory Map and Allocations:
;
; $000-$5ff (user code)      1536 bytes
; $600-$6f9 (<record> data)   250 bytes
; $6fa-$6ff (wasted)            6 bytes
; $700-$7df (globals)         224 bytes
; $7e0/$7e1 (button #1 ptr)     2 bytes
; $7e2/$7e3 (button #2 ptr)     2 bytes
; $7e4-$7ef (wasted)           12 bytes
; $7f0-$7ff (cricket name)     16 bytes
;
; I am using $000-$4ff (1280 bytes) for partition 1: downloaded (p1 and/or p2) code
;            $500-$5ff ( 256 bytes) for partition 2: code to be immediately downloaded and run
;            $7e4-$7e6 for the p1 proc call ($7e0/$7e1 is set to point here)
;            $7e7-$7e9 for the p2 proc call ($7e2/$7e3 is set to point here)

;---------------

; Blue Dot Cricket EEPROM Memory Map, Allocations, and Notes:
;
; $000-$4ff (user code)      1280 bytes
; $500-$ec3 (<record> data)  2500 bytes
; $ec4-$fef (user code)       300 bytes
; $ff0/$ff1 (button #1 ptr)     2 bytes
; $ff2/$ff3 (button #2 ptr)     2 bytes
; $ff4-$ffe (cricket name)     11 bytes
; $fff      (autostart flag)    1 bytes
;
; I am using $000-$4ff (1280 bytes) for partition 1: downloaded (p1 and/or p2) code
;            $ec4-$fef  (300 bytes) for partition 2: code to be immediately downloaded and run
;            $ff4-$ff6 for the p1 proc call ($ff0/$ff1 is set to point here)
;            $ff7-$ff9 for the p2 proc call ($ff2/$ff3 is set to point here)
;
; Notes: blue dots support user-defined arrays; they store 16-bit numbers and start at $000
;        blue dots support 16 global variables; they are 16-bit numbers and are stored in the cricket's RAM
;        neither are used by LogoBlocks in blue-dot mode

;---------------

(defmethod sufficient-partition1-memory? ((view wrkarea-view))
  (let ((curpos #x000))
    (loop for block in (masterlist view) do
          (if (equal (species block) 'when)
            (setf curpos (+ curpos (length (compile-from-block view block))))))
    (loop for block in (masterlist view) do
          (if (equal (species block) 'defproc)
            (progn (setf (mempos block) curpos)
                   (setf curpos (+ curpos (length (compile-from-block view block)))))))
    (if (> curpos #x500) nil t)))

;--------------- DOWNLOAD OPS ------------------------------------

(defmethod download-all-proc-clusters ((view wrkarea-view))
  (let ((p1-call-address (cond ((= *active-project-mode* 1) #x7e4)
                               ((= *active-project-mode* 2) #xff4)))
        (p2-call-address (cond ((= *active-project-mode* 1) #x7e7)
                               ((= *active-project-mode* 2) #xff7)))
        (p1-call-data    (append (callproc-bytecodes view 1) (list 0)))
        (p2-call-data    (append (if (index-defproc-block view 2)
                                   (callproc-bytecodes view 2)) (list 0)))
        (result 0))
    (loop for block in (masterlist view) do
          (if (= result 0)
            (if (equal (species block) 'when)
              (setf result (download #x000 (compile-from-block view block) t nil)))))
    (loop for block in (masterlist view) do
          (if (= result 0)
            (if (equal (species block) 'defproc)
              (setf result (download (mempos block) (compile-from-block view block) t nil)))))
    (if (= result 0) (setf result (set-first-ir-vector  p1-call-address)))
    (if (= result 0) (setf result (set-second-ir-vector p2-call-address)))
    (if (= result 0) (setf result (download p1-call-address p1-call-data nil nil)))
    (if (= result 0) (setf result (download p2-call-address p2-call-data nil t)))
    result))

;--------------- BLOCK COMPILERS ------------------------------------

(defmethod compile-from-block ((view wrkarea-view) (block on/off-block))
  (let ((block1 (scan-dock view block 1)))
    (append (motor-bytecode block)
            (cond ((= (command block) 1) (trbc '<on>   ))
                  ((= (command block) 2) (trbc '<off>  ))
                  ((= (command block) 3) (trbc '<brake>)))
            (if block1 (compile-from-block view block1)))))

(defmethod compile-from-block ((view wrkarea-view) (block direction-block))
  (let ((block1 (scan-dock view block 1)))
    (append (motor-bytecode block)
            (cond ((= (command block) 1) (trbc '<thisway>))
                  ((= (command block) 2) (trbc '<thatway>))
                  ((= (command block) 3) (trbc '<rd>     )))
            (if block1 (compile-from-block view block1)))))

(defmethod compile-from-block ((view wrkarea-view) (block power-block))
  (let ((block1 (scan-dock view block 1)))
    (append (motor-bytecode block)
            (trnm (command block))
            (trbc '<setpower>)
            (if block1 (compile-from-block view block1)))))

(defmethod compile-from-block ((view wrkarea-view) (block onfor-block))
  (let ((block1 (scan-dock view block 1))
        (block2 (scan-dock view block 2)))
    (append (motor-bytecode block)
            (compile-from-block view block1)
            (trbc '<onfor>)
            (if block2 (compile-from-block view block2)))))

(defmethod compile-from-block ((view wrkarea-view) (block beep-block))
  (let ((block1 (scan-dock view block 1)))
    (append (trbc '<beep>)
            (if block1 (compile-from-block view block1)))))

(defmethod compile-from-block ((view wrkarea-view) (block note-block))
  (let ((block1 (scan-dock view block 1))
        (block2 (scan-dock view block 2)))
    (append (music-bytecode block)
            (compile-from-block view block1)
            (trbc '<note>)
            (if block2 (compile-from-block view block2)))))

(defmethod compile-from-block ((view wrkarea-view) (block wait-block))
  (let ((block1 (scan-dock view block 1))
        (block2 (scan-dock view block 2)))
    (append (compile-from-block view block1)
            (trbc '<wait>)
            (if block2 (compile-from-block view block2)))))

(defmethod compile-from-block ((view wrkarea-view) (block zerotimer-block))
  (let ((block1 (scan-dock view block 1)))
    (append (trbc '<resett>)
            (if block1 (compile-from-block view block1)))))

(defmethod compile-from-block ((view wrkarea-view) (block timer-block))
  (trbc '<timer>))

(defmethod compile-from-block ((view wrkarea-view) (block dummy-block))
  (let ((block1 (scan-dock view block 1)))
    (append (if block1 (compile-from-block view block1)))))

(defmethod compile-from-block ((view wrkarea-view) (block when-block))
  (let ((block0 (scan-dock view block 0))
        (block1 (scan-dock view block 1)))
    (append (list 0)
            (if block1
              (append (convert-to-eolr (compile-from-block view block0))
                      (convert-to-list (compile-from-block view block1))
                      (trbc '<when>)))
            (trbc '<stop>))))

(defmethod compile-from-block ((view wrkarea-view) (block whenoff-block))
  (let ((block1 (scan-dock view block 1)))
    (append (trbc '<whenoff>)
            (if block1 (compile-from-block view block1)))))

(defmethod compile-from-block ((view wrkarea-view) (block if-block))
  (let ((block1 (scan-dock view block 1))
        (block2 (scan-dock view block 2))
        (block3 (scan-dock view block 3)))
    (append (compile-from-block view block1)
            (convert-to-list (compile-from-block view block2))
            (trbc '<if>)
            (if block3 (compile-from-block view block3)))))

(defmethod compile-from-block ((view wrkarea-view) (block ifelse-block))
  (let ((block1 (scan-dock view block 1))
        (block2 (scan-dock view block 2))
        (block3 (scan-dock view block 3))
        (block4 (scan-dock view block 4)))
    (append (compile-from-block view block1)
            (convert-to-list (compile-from-block view block2))
            (convert-to-list (compile-from-block view block3))
            (trbc '<ifelse>)
            (if block4 (compile-from-block view block4)))))

(defmethod compile-from-block ((view wrkarea-view) (block repeat-block))
  (let ((block1 (scan-dock view block 1))
        (block2 (scan-dock view block 2))
        (block3 (scan-dock view block 3)))
    (append (compile-from-block view block1)
            (convert-to-list (compile-from-block view block2))
            (trbc '<repeat>)
            (if block3 (compile-from-block view block3)))))

(defmethod compile-from-block ((view wrkarea-view) (block loop-block))
  (let ((block1 (scan-dock view block 1))
        (block2 (scan-dock view block 2)))
    (append (convert-to-list (compile-from-block view block1))
            (trbc '<loop>)
            (if block2 (compile-from-block view block2)))))

(defmethod compile-from-block ((view wrkarea-view) (block waituntil-block))
  (let ((block1 (scan-dock view block 1))
        (block2 (scan-dock view block 2)))
    (append (convert-to-eolr (compile-from-block view block1))
            (trbc '<waituntil>)
            (if block2 (compile-from-block view block2)))))

(defmethod compile-from-block ((view wrkarea-view) (block switch-block))
  (let ((block1 (scan-dock view block 1)))
    (append (cond ((= (switch block) 1) (trbc '<switcha>))
                  ((= (switch block) 2) (trbc '<switchb>)))
            (if block1 (compile-from-block view block1)))))

(defmethod compile-from-block ((view wrkarea-view) (block sensor-block))
  (cond ((= (sensor block) 1) (trbc '<sensora>))
        ((= (sensor block) 2) (trbc '<sensorb>))))

; global [cached-ir-value use-cached-value?]
;
; to clearir-macro
;   setuse-cached-value? 0
; end

(defmethod compile-from-block ((view wrkarea-view) (block clearir-block))
  (let ((block1 (scan-dock view block 1)))
    (append (trnm 1)
            (trnm 0)
            (trbc '<db>)
            (if block1 (compile-from-block view block1)))))

(defmethod compile-from-block ((view wrkarea-view) (block newir?-block))
  (let ((block1 (scan-dock view block 1)))
    (append (trbc '<newir?>)
            (if block1 (compile-from-block view block1)))))

; to ir-macro
;   if use-cached-value? = 0 [
;     setcached-ir-value ir
;     setuse-cached-value? 1]
;   output cached-ir-value
; end

(defmethod compile-from-block ((view wrkarea-view) (block ir-block))
  (cond ((= *active-project-mode* 1) (trnm 1)
                                     (trbc '<eb>)
                                     (trnm 0)
                                     (trbc '<=>)
                                     (convert-to-list (append (trnm 0)
                                                              (trbc '<ir>)
                                                              (trbc '<db>)
                                                              (trnm 1)
                                                              (trnm 1)
                                                              (trbc '<db>)))
                                     (trbc '<if>)
                                     (trnm 0)
                                     (trbc '<eb>))
        ((= *active-project-mode* 2) (trbc '<ir>))))

(defmethod compile-from-block ((view wrkarea-view) (block sendir-block))
  (let ((block1 (scan-dock view block 1))
        (block2 (scan-dock view block 2)))
    (append (compile-from-block view block1)
            (trbc '<send>)
            (if block2 (compile-from-block view block2)))))

(defmethod compile-from-block ((view wrkarea-view) (block display-block))
  (let ((block1 (scan-dock view block 1))
        (block2 (scan-dock view block 2)))
    (append (compile-from-block view block1)
            (trnm 0)
            (cond ((= (color block) 1) (trnm #x14c))
                  ((= (color block) 2) (trnm #x14d))
                  ((= (color block) 3) (trnm #x14e))
                  ((= (color block) 4) (trnm #x14f)))
            (trbc '<bsend>)
            (trbc '<bsend>)
            (trbc '<bsend>)
            (if block2 (compile-from-block view block2)))))

(defmethod compile-from-block ((view wrkarea-view) (block number-block))
  (cond ((= (genmethod block) 1) (trnm (value block)))
        ((= (genmethod block) 2) (append (trbc '<random>)
                                         (trnm (- (+ (upperbound block) 1) (lowerbound block)))
                                         (trbc '<%>)
                                         (trnm (lowerbound block))
                                         (trbc '<+>)))))

(defmethod compile-from-block ((view wrkarea-view) (block boolean-block))
  (let ((block1 (scan-dock view block 1))
        (block2 (scan-dock view block 2))
        (block3 (scan-dock view block 3)))
    (append (compile-from-block view block1)
            (compile-from-block view block2)
            (cond ((= (comparator block) 1) (trbc '<<>))
                  ((= (comparator block) 2) (trbc '<=>))
                  ((= (comparator block) 3) (trbc '<>>)))
            (if block3 (compile-from-block view block3)))))

(defmethod compile-from-block ((view wrkarea-view) (block and-block))
  (let ((block1 (scan-dock view block 1)))
    (append (compile-from-block view block1)
            (trbc '<and>))))

(defmethod compile-from-block ((view wrkarea-view) (block or-block))
  (let ((block1 (scan-dock view block 1)))
    (append (compile-from-block view block1)
            (trbc '<or>))))

(defmethod compile-from-block ((view wrkarea-view) (block cleardata-block))
  (let ((block1 (scan-dock view block 1)))
    (append (cond ((= *active-project-mode* 1) (trbc '<resetdp>)
                                               (trnm 250)
                                               (convert-to-list (append (trnm 0)
                                                                        (trbc '<record>)))
                                               (trbc '<repeat>)
                                               (trbc '<resetdp>))
                  ((= *active-project-mode* 2) (trbc '<erase>)))
            (if block1 (compile-from-block view block1)))))

(defmethod compile-from-block ((view wrkarea-view) (block record-block))
  (let ((block1 (scan-dock view block 1))
        (block2 (scan-dock view block 2)))
    (append (compile-from-block view block1)
            (trbc '<record>)
            (if block2 (compile-from-block view block2)))))

(defmethod compile-from-block ((view wrkarea-view) (block resetdp-block))
  (let ((block1 (scan-dock view block 1)))
    (append (trbc '<resetdp>)
            (if block1 (compile-from-block view block1)))))

(defmethod compile-from-block ((view wrkarea-view) (block defproc-block))
  (let ((when?  (and (< (procnum block) 3)
                     (index-when-block view)))
        (block0 (scan-dock view block 0)))
    (append (list 0)
            (if when? (list 128 0))
            (if block0 (compile-from-block view block0))
            (trbc '<stop>))))

(defmethod compile-from-block ((view wrkarea-view) (block stop-block))
  (trbc '<stop>))

(defmethod compile-from-block ((view wrkarea-view) (block callproc-block))
  (let ((block1 (scan-dock view block 1)))
    (append (if (> (procnum block) 0) (callproc-bytecodes view (procnum block)))
            (if block1 (compile-from-block view block1)))))

;--------------- BYTECODE OPS ------------------------------------

(defun motor-bytecode (block)
  (trbc (if (motorA? block) (if (motorB? block) '<ab> '<a>) '<b>)))

(defun music-bytecode (block)
  (let ((12th-root-of-2  1.05946)
        (base-frequency  440)
        (octave-modifier (cond ((= (octave block) 1)  12)
                               ((= (octave block) 2)   0)
                               ((= (octave block) 3) -12)))
        (pitch-modifier  (cond ((= (pitch block) 1)  0)
                               ((= (pitch block) 2)  2)
                               ((= (pitch block) 3)  3)
                               ((= (pitch block) 4)  5)
                               ((= (pitch block) 5)  7)
                               ((= (pitch block) 6)  8)
                               ((= (pitch block) 7) 10))))
    (setf pitch-modifier (+ pitch-modifier (cond ((= (modifier block) 1)  1)
                                                 ((= (modifier block) 2)  0)
                                                 ((= (modifier block) 3) -1))))
    (setf base-frequency (* base-frequency (x-to-the-y 12th-root-of-2 (+ octave-modifier pitch-modifier))))
    (trnm (round 40000 base-frequency))))

(defmethod callproc-bytecodes ((view wrkarea-view) procnum)
  (let ((block (index-defproc-block view procnum)))
    (if (mempos block) (list (+ (ldb (byte 8 8) (mempos block)) 128)    ; +128 sets the high bit
                             (+ (ldb (byte 8 0) (mempos block))   0))
                       (list 0 0))))

(defun convert-to-list (bytelist)
  (append (trbc '<%list>)
          (list (+ (length bytelist) 1))
          bytelist
          (trbc '<%eol>)))

(defun convert-to-eolr (bytelist)
  (append (trbc '<%list>)
          (list (+ (length bytelist) 1))
          bytelist
          (trbc '<%eolr>)))

;---------------

(defun trbc (bc)
  (list (position bc (cond ((= *active-project-mode* 1) '(nil <%num> <%list> <%eol> <%eolr> <%input> <stop> <output> <db> <eb>
                                                          <repeat> <loop> <waituntil> <if> <ifelse> <beep> <note> <wait>
                                                          <+> <-> <*> </> <%> <=> <>> <<> <and> <or> <not>
                                                          <send> <ir> <a> <b> <ab> <on> <onfor> <off> <thisway> <thatway> <rd>
                                                          <sensora> <sensorb> <switcha> <switchb> <random> <timer> <resett> <when>
                                                          <off#> <red#> <green#> <amber#> <red> <green> <amber>
                                                          <alloff> <show> <sensorc> <switchc> <setpower> <brake> <record>
                                                          <recall> <setdp> <resetdp>))
                           ((= *active-project-mode* 2) '(nil <%num> <%num16> <%list> <%eol> <%eolr> <%input> <stop> <output>
                                                          <repeat> <if> <ifelse> <beep> <note> <waituntil> <loop>
                                                          <wait> <timer> <resett> <send> <ir> <newir?> <random>
                                                          <+> <-> <*> </> <%> <=> <>> <<> <and> <or> <xor> <not> <db> <eb>
                                                          <aset> <aget> <record> <recall> <resetdp> <setdp> <erase> <when> <whenoff>
                                                          <a> <b> <ab> <on> <onfor> <off> <thisway> <thatway> <rd>
                                                          <sensora> <sensorb> <switcha> <switchb>
                                                          <setpower> <brake> <bsend> <bsr> <c> <d> <cd> <abcd>))))))

(defun trnm (nm)
  (if (< nm 256)
    (append (trbc '<%num>)   (list nm))
    (append (trbc '<%num16>) (list (ldb (byte 8 8) nm)
                                   (ldb (byte 8 0) nm)))))