;--------------- PREFERENCES OPS ------------------------------------

(defun read-preferences ()
  (if (directory "CCL:Preferences")
    (progn (with-open-file (input-stream "CCL:Preferences" :direction :input)
             (read-line input-stream)
             (setf *default-project-mode*  (read-from-string (read-line input-stream)))
             (setf  *active-project-mode*  *default-project-mode*)
             (setf *interface-serial-port* 1)
             (setf *default-window-size*   (read-from-string (read-line input-stream)))
             (setf *current-window-size*   *default-window-size*)
             (setf *size-borders-flag*     (read-from-string (read-line input-stream))))
           (set-mac-file-type    "CCL:Preferences" :|PREF|)
           (set-mac-file-creator "CCL:Preferences" :|LgBl|))
    (progn (setf *default-project-mode*  1)
           (setf  *active-project-mode*  *default-project-mode*)
           (setf *interface-serial-port* 1)
           (setf *default-window-size*   1)
           (setf *current-window-size*   *default-window-size*)
           (setf *size-borders-flag*     0)
           (write-preferences))))

(defun write-preferences ()
  (with-open-file (output-stream "CCL:Preferences" :direction :output
                                                   :if-exists :supersede)
    (print *default-project-mode* output-stream)
    (print *default-window-size*  output-stream)
    (print *size-borders-flag*    output-stream))
  (set-mac-file-type    "CCL:Preferences" :|PREF|)
  (set-mac-file-creator "CCL:Preferences" :|LgBl|))

;--------------- DOCUMENT OPS ------------------------------------

(defmethod new-file ((window logoblocks-window))
  (if (save?-dialog window)
    (with-cursor *watch-cursor*
      (setf *file-path* nil)
      (setf *active-project-mode* *default-project-mode*)
      (update-window-title)
      (delete-all-blocks  (wrkarea   window))
      (reset-view         (scrollbar window) nil)
      (redo-all-blocks    (palette   window))
      (rebuild-all-blocks (wrkarea   window))
      (clear-flag))))

(defmethod open-file ((window logoblocks-window) &optional (path nil))
  (if (save?-dialog window)
    (progn (if (not path) (setf path (catch-cancel (choose-file-dialog :mac-file-type :|PRJC|))))
           (unless (eq path :cancel)
             (with-cursor *watch-cursor*
               (setf *file-path* path)
               (let ((prjcmode (with-open-file (input-stream *file-path* :direction :input)
                                 (read-line input-stream)
                                 (read-line input-stream)))
                     (filedata (with-open-file (input-stream *file-path* :direction :input)
                                 (read-line input-stream)
                                 (read-line input-stream)
                                 (loop for i = (peek-char t input-stream nil 'eof)
                                       until (or (equal i #\@)
                                                 (equal i 'eof))
                                       collect (loop for i = (read-line input-stream nil 'eof)
                                                     until (or (equal i "@#$#@#$#@")
                                                               (equal i "")
                                                               (equal i 'eof))
                                                     collect i)))))
                 (set-mac-file-type    *file-path* :|PRJC|)
                 (set-mac-file-creator *file-path* :|LgBl|)
                 (setf *active-project-mode* (read-from-string prjcmode))
                 (update-window-title)
                 (delete-all-blocks (wrkarea   window))
                 (reset-view        (scrollbar window) nil)
                 (redo-all-blocks   (palette   window))
                 (restore-blocks    (wrkarea   window) filedata)
                 (clear-flag)))))))

(defmethod save-file ((window logoblocks-window))
  (if (not *file-path*)
    (save-file-as window)
    (with-cursor *watch-cursor*
      (with-open-file (output-stream *file-path* :direction :output
                                                 :if-exists :supersede)
        (print *active-project-mode* output-stream)
        (loop for blockdata in (save-blocks (wrkarea window)) do
              (terpri output-stream)
              (loop for slot in blockdata do
                    (print slot output-stream)))
        (print "@#$#@#$#@" output-stream))
      (set-mac-file-type    *file-path* :|PRJC|)
      (set-mac-file-creator *file-path* :|LgBl|)
      (clear-flag))))

(defmethod save-file-as ((window logoblocks-window))
  (let ((path (catch-cancel (choose-new-file-dialog :prompt "Save As"))))
    (unless (eq path :cancel)
      (setf *file-path* path)
      (update-window-title)
      (save-file window))))

;--------------- BLOCK STORAGE OPS ------------------------------------

(defmethod save-blocks ((view wrkarea-view))
  (let ((result ()))
    (loop for block in (masterlist view) do
          (setf result (append result (list (append (list (left      (boundrect block))
                                                          (top       (boundrect block))
                                                          (species   block)
                                                          (selected? block))
                                                    (save-slots block))))))
    result))

;---------------

(defmethod save-slots ((block on/off-block))
  (list (command block)
        (motorA? block)
        (motorB? block)))

(defmethod save-slots ((block direction-block))
  (list (command block)
        (motorA? block)
        (motorB? block)))

(defmethod save-slots ((block power-block))
  (list (command block)
        (motorA? block)
        (motorB? block)))

(defmethod save-slots ((block onfor-block))
  (list (motorA? block)
        (motorB? block)))

(defmethod save-slots ((block beep-block))
  (list))

(defmethod save-slots ((block note-block))
  (list (pitch    block)
        (modifier block)
        (octave   block)))

(defmethod save-slots ((block wait-block))
  (list))

(defmethod save-slots ((block zerotimer-block))
  (list))

(defmethod save-slots ((block timer-block))
  (list))

(defmethod save-slots ((block dummy-block))
  (list))

(defmethod save-slots ((block when-block))
  (list))

(defmethod save-slots ((block whenoff-block))
  (list))

(defmethod save-slots ((block if-block))
  (list))

(defmethod save-slots ((block ifelse-block))
  (list))

(defmethod save-slots ((block repeat-block))
  (list))

(defmethod save-slots ((block loop-block))
  (list))

(defmethod save-slots ((block waituntil-block))
  (list))

(defmethod save-slots ((block switch-block))
  (list (switch block)))

(defmethod save-slots ((block sensor-block))
  (list (sensor block)))

(defmethod save-slots ((block clearir-block))
  (list))

(defmethod save-slots ((block newir?-block))
  (list))

(defmethod save-slots ((block ir-block))
  (list))

(defmethod save-slots ((block sendir-block))
  (list))

(defmethod save-slots ((block display-block))
  (list (color block)))

(defmethod save-slots ((block number-block))
  (list (genmethod  block)
        (value      block)
        (lowerbound block)
        (upperbound block)))

(defmethod save-slots ((block boolean-block))
  (list (comparator block)))

(defmethod save-slots ((block and-block))
  (list))

(defmethod save-slots ((block or-block))
  (list))

(defmethod save-slots ((block cleardata-block))
  (list))

(defmethod save-slots ((block record-block))
  (list))

(defmethod save-slots ((block resetdp-block))
  (list))

(defmethod save-slots ((block defproc-block))
  (list (procnum   block)
        (proctitle block)))

(defmethod save-slots ((block stop-block))
  (list))

(defmethod save-slots ((block callproc-block))
  (list (procnum   block)
        (proctitle block)))

;--------------- BLOCK RESTORATION OPS ------------------------------------

(defmethod restore-blocks ((view wrkarea-view) filedata)
  (loop for scan in filedata do
        (let* ((left      (read-from-string (nth 0 scan)))
               (top       (read-from-string (nth 1 scan)))
               (species   (read-from-string (nth 2 scan)))
               (selected? (read-from-string (nth 3 scan)))
               (block     (create-block view species t selected? (make-point left top))))
          (restore-slots block scan)))
  (draw-buffer3       view (masterlist view) 'draw)
  (view-draw-contents view))

;---------------

(defmethod restore-slots ((block on/off-block) scan)
  (setf (command block) (read-from-string (nth 4 scan)))
  (setf (motorA? block) (read-from-string (nth 5 scan)))
  (setf (motorB? block) (read-from-string (nth 6 scan))))

(defmethod restore-slots ((block direction-block) scan)
  (setf (command block) (read-from-string (nth 4 scan)))
  (setf (motorA? block) (read-from-string (nth 5 scan)))
  (setf (motorB? block) (read-from-string (nth 6 scan))))

(defmethod restore-slots ((block power-block) scan)
  (setf (command block) (read-from-string (nth 4 scan)))
  (setf (motorA? block) (read-from-string (nth 5 scan)))
  (setf (motorB? block) (read-from-string (nth 6 scan))))

(defmethod restore-slots ((block onfor-block) scan)
  (setf (motorA? block) (read-from-string (nth 4 scan)))
  (setf (motorB? block) (read-from-string (nth 5 scan))))

(defmethod restore-slots ((block beep-block) scan)
  (declare (ignore block scan)))

(defmethod restore-slots ((block note-block) scan)
  (setf (pitch    block) (read-from-string (nth 4 scan)))
  (setf (modifier block) (read-from-string (nth 5 scan)))
  (setf (octave   block) (read-from-string (nth 6 scan))))

(defmethod restore-slots ((block wait-block) scan)
  (declare (ignore block scan)))

(defmethod restore-slots ((block zerotimer-block) scan)
  (declare (ignore block scan)))

(defmethod restore-slots ((block timer-block) scan)
  (declare (ignore block scan)))

(defmethod restore-slots ((block dummy-block) scan)
  (declare (ignore block scan)))

(defmethod restore-slots ((block when-block) scan)
  (declare (ignore block scan)))

(defmethod restore-slots ((block whenoff-block) scan)
  (declare (ignore block scan)))

(defmethod restore-slots ((block if-block) scan)
  (declare (ignore block scan)))

(defmethod restore-slots ((block ifelse-block) scan)
  (declare (ignore block scan)))

(defmethod restore-slots ((block repeat-block) scan)
  (declare (ignore block scan)))

(defmethod restore-slots ((block loop-block) scan)
  (declare (ignore block scan)))

(defmethod restore-slots ((block waituntil-block) scan)
  (declare (ignore block scan)))

(defmethod restore-slots ((block switch-block) scan)
  (setf (switch block) (read-from-string (nth 4 scan))))

(defmethod restore-slots ((block sensor-block) scan)
  (setf (sensor block) (read-from-string (nth 4 scan))))

(defmethod restore-slots ((block clearir-block) scan)
  (declare (ignore block scan)))

(defmethod restore-slots ((block newir?-block) scan)
  (declare (ignore block scan)))

(defmethod restore-slots ((block ir-block) scan)
  (declare (ignore block scan)))

(defmethod restore-slots ((block sendir-block) scan)
  (declare (ignore block scan)))

(defmethod restore-slots ((block display-block) scan)
  (setf (color block) (read-from-string (nth 4 scan))))

(defmethod restore-slots ((block number-block) scan)
  (setf (genmethod  block) (read-from-string (nth 4 scan)))
  (setf (value      block) (read-from-string (nth 5 scan)))
  (setf (lowerbound block) (read-from-string (nth 6 scan)))
  (setf (upperbound block) (read-from-string (nth 7 scan))))

(defmethod restore-slots ((block boolean-block) scan)
  (setf (comparator block) (read-from-string (nth 4 scan))))

(defmethod restore-slots ((block and-block) scan)
  (declare (ignore block scan)))

(defmethod restore-slots ((block or-block) scan)
  (declare (ignore block scan)))

(defmethod restore-slots ((block cleardata-block) scan)
  (declare (ignore block scan)))

(defmethod restore-slots ((block record-block) scan)
  (declare (ignore block scan)))

(defmethod restore-slots ((block resetdp-block) scan)
  (declare (ignore block scan)))

(defmethod restore-slots ((block defproc-block) scan)
  (setf (procnum   block) (read-from-string (nth 4 scan)))
  (setf (proctitle block) (read-from-string (nth 5 scan))))

(defmethod restore-slots ((block stop-block) scan)
  (declare (ignore block scan)))

(defmethod restore-slots ((block callproc-block) scan)
  (setf (procnum   block) (read-from-string (nth 4 scan)))
  (setf (proctitle block) (read-from-string (nth 5 scan))))