;--------------- WINDOW-LEVEL BUFFER3 OPS ------------------------------------

(defmethod draw-buffer3 ((view buttons-view) blocklist blockcommand)
  (declare (ignore blocklist blockcommand))
  (do-in-gworld (buffer3 (view-container view))
                (with-back-color *light-gray-color*
                  (#_EraseRect (globalrect view)))
                (with-fore-color *black-color*
                  (#_FrameRect (globalrect view)))
                (loop for x in (first-n-integers 5) do
                      (draw-button view x (if (= (switch view) x) t nil)))))

(defmethod draw-buffer3 ((view working-view) blocklist blockcommand)
  (do-in-gworld (buffer3 (view-container view))
                (with-back-color (if (eq view (palette (view-container view))) *light-gray-color* *white-color*)
                  (#_EraseRect (globalrect view)))
                (with-fore-color *black-color*
                  (#_FrameRect (globalrect view)))
                (if (eq view (wrkarea (view-container view)))
                  (draw-markers view)))
  (loop for block in blocklist do
        (if (visible? block)
          (cond ((equal blockcommand 'draw)  (draw-shadow  view block 1)
                                             (draw-block   view block 2))
                ((equal blockcommand 'erase) (erase-shadow view block)
                                             (erase-block  view block)))))
  (gworld-maskthru (buffer1                 view)
                   (buffer1                 view)
                   (buffer3 (view-container view))
                   ( localrect view)
                   ( localrect view)
                   (offsetrect view))
  (gworld-maskthru (buffer2                 view)
                   (buffer1                 view)
                   (buffer3 (view-container view))
                   ( localrect view)
                   ( localrect view)
                   (globalrect view)))

;--------------- BUTTONS-VIEW GRAPHICS ------------------------------------

(defmethod draw-button ((view buttons-view) button enabled?)
  (let* ((origin   (view-position view))
         (oinverse (make-point (- (point-h origin))
                               (- (point-v origin))))
         (rect     (aref (controls view) button))
         (title    (cond ((= button 0) "Action")
                         ((= button 1) "Control")
                         ((= button 2) "Sensor")
                         ((= button 3) "Procedure")))
         (gray1    (make-color 15000 15000 15000))
         (gray2    (make-color 22500 22500 22500))
         (gray3    (make-color 32500 32500 32500))
         (gray4    (make-color 40000 40000 40000))
         (gray5    (make-color 47500 47500 47500))
         (gray6    (make-color 55000 55000 55000))
         (gray7    (make-color 65000 65000 65000)))
    (offset-rect rect origin)
    (with-fore-color gray4
      (#_PaintRect rect))
    (draw-border rect enabled? *black-color* *black-color*)
    (#_InsetRect rect  1  1)
    (draw-border rect enabled? gray1 gray7)
    (#_InsetRect rect  1  1)
    (draw-border rect enabled? gray2 gray6)
    (#_InsetRect rect  1  1)
    (draw-border rect enabled? gray3 gray5)
    (#_InsetRect rect -3 -3)
    (with-fore-color (if (= button 4) gray2 (if enabled? *blue-color* *black-color*))
      (multiple-value-bind (ff ms) (font-codes '("Helvetica" 18))
        (set-wptr-font-codes (buffer3 (view-container view)) ff ms)
        (if (< button 4)
          (progn (#_MoveTo (- (point-h (rect-centre rect))
                              (truncate (font-codes-string-width title ff ms) 2))
                           (+ (point-v (rect-centre rect))
                              (truncate (font-codes-line-height        ff ms) 2)
                              -2))
                 (with-pstrs ((pstr title))
                   (#_DrawString pstr)))
          (progn (#_MoveTo (- (point-h (rect-centre rect))
                              (truncate (font-codes-string-width "M" ff ms) 2))
                           (+ (point-v (rect-centre rect)) -21 6))
                 (with-pstrs ((pstr "M")) (#_DrawString pstr))
                 (#_MoveTo (- (point-h (rect-centre rect))
                              (truncate (font-codes-string-width "i" ff ms) 2))
                           (+ (point-v (rect-centre rect))  -5 6))
                 (with-pstrs ((pstr "i")) (#_DrawString pstr))
                 (#_MoveTo (- (point-h (rect-centre rect))
                              (truncate (font-codes-string-width "s" ff ms) 2))
                           (+ (point-v (rect-centre rect))  10 6))
                 (with-pstrs ((pstr "s")) (#_DrawString pstr))
                 (#_MoveTo (- (point-h (rect-centre rect))
                              (truncate (font-codes-string-width "c" ff ms) 2))
                           (+ (point-v (rect-centre rect))  23 6))
                 (with-pstrs ((pstr "c")) (#_DrawString pstr))))))
    (offset-rect rect oinverse)))

;---------------

(defun draw-border (rect enabled? color1 color2)
  (with-fore-color (if enabled? color1 color2)
    (#_MoveTo (left  rect) (top    rect))
    (#_LineTo (right rect) (top    rect))
    (#_MoveTo (left  rect) (top    rect))
    (#_LineTo (left  rect) (bottom rect)))
  (with-fore-color (if enabled? color2 color1)
    (#_MoveTo (right rect) (bottom rect))
    (#_LineTo (left  rect) (bottom rect))
    (#_MoveTo (right rect) (bottom rect))
    (#_LineTo (right rect) (top    rect))))

;--------------- WRKAREA-VIEW GRAPHICS ------------------------------------

(defmethod draw-markers ((view wrkarea-view))
  (unless (= *size-borders-flag* 0)
    (multiple-value-bind (ff ms) (font-codes '("Monaco" 9))
      (set-wptr-font-codes (buffer3 (view-container view)) ff ms))
    (with-fore-color *light-blue-color*
      (if (>= *current-window-size* 2)
        (progn (#_MoveTo 620   0)
               (#_LineTo 620 425)
               (#_LineTo 126 425)
               (let ((linetext "640 x 480"))
                 (#_MoveTo 564 422)
                 (with-pstrs ((pstr linetext))
                   (#_DrawString pstr)))))
      (if (>= *current-window-size* 3)
        (progn (#_MoveTo 780   0)
               (#_LineTo 780 545)
               (#_LineTo 126 545)
               (let ((linetext "800 x 600"))
                 (#_MoveTo 724 542)
                 (with-pstrs ((pstr linetext))
                   (#_DrawString pstr))))))))

;--------------- BLOCK GRAPHICS ------------------------------------

(defmethod erase-shadow ((view working-view) block)
  (do-in-gworld (buffer1 view)
                (with-fore-color *white-color*
                  (#_PaintRgn (boundrgn block)))))

(defmethod erase-block ((view working-view) block)
  (do-in-gworld (buffer2 view)
                (with-fore-color *white-color*
                  (#_PaintRgn (boundrgn block)))))

(defmethod draw-shadow ((view working-view) block buffer)
  (let ((target-buffer (cond ((= buffer 1) (buffer1                 view))
                             ((= buffer 4) (buffer4 (view-container view))))))
    (do-in-gworld target-buffer
                  (with-fore-color *black-color*
                    (#_PaintRgn (boundrgn block))))))

(defmethod draw-block ((view working-view) block buffer)
  (let ((pict-buffer   (nth (+ (* (family block) 3) (if (selected? block) 1 0)) *pict-buffers*))
        (target-buffer (cond ((= buffer 2) (buffer2                 view))
                             ((= buffer 4) (buffer4 (view-container view))))))
    (gworld-copythru   pict-buffer
                     target-buffer
                     (access-gworld-boundrect pict-buffer)
                     (boundrect block)
                     (boundrgn  block))
    (do-in-gworld    target-buffer
                     (draw-block-text block))
    (if (and (equal (species block) 'direction)
             (fulltext? block)
             (< (command block) 3))
      (let* ((source-rect (access-gworld-boundrect (nth 72 *pict-buffers*)))
             (offset      (add-points (topleft (boundrect block))
                                      (make-point (+ (truncate (- (rect-width (boundrect block))
                                                                  (rect-width source-rect))
                                                               2)
                                                     10)
                                                  (truncate (- (rect-height (boundrect block))
                                                               (rect-height source-rect))
                                                            2)))))
        (rlet ((target-rect :rect :topleft  (add-points (topleft  source-rect) offset)
                                  :botright (add-points (botright source-rect) offset)))
          (gworld-maskthru (nth 72 *pict-buffers*)
                           (nth (+ 72 (command block)) *pict-buffers*)
                           target-buffer
                           source-rect
                           source-rect
                           target-rect))))))

;---------------

(defmethod draw-block-text ((block separator-block))
  ())

(defmethod draw-block-text ((block on/off-block))
  (with-fore-color *white-color*
    (let ((blocktext (if (fulltext? block)
                       (concatenate 'string (if (motorA? block) "A")
                                            (if (motorB? block) "B")
                                            (cond ((= (command block) 1) " ON")
                                                  ((= (command block) 2) " OFF")
                                                  ((= (command block) 3) " BRAKE")))
                       "ON/OFF")))
      (#_MoveTo (- (point-h (rect-centre (boundrect block))) (truncate (string-width blocktext) 2))
                (+ (point-v (rect-centre (boundrect block))) 3))
      (with-pstrs ((pstr blocktext))
        (#_DrawString pstr)))))

(defmethod draw-block-text ((block direction-block))
  (with-fore-color *white-color*
    (let ((blocktext (if (fulltext? block)
                       (concatenate 'string (if (motorA? block) "A")
                                            (if (motorB? block) "B")
                                            (cond ((= (command block) 1) "      ")
                                                  ((= (command block) 2) "      ")
                                                  ((= (command block) 3) " RD")))
                       "Direction")))
      (#_MoveTo (- (point-h (rect-centre (boundrect block))) (truncate (string-width blocktext) 2))
                (+ (point-v (rect-centre (boundrect block))) 3))
      (with-pstrs ((pstr blocktext))
        (#_DrawString pstr)))))

(defmethod draw-block-text ((block power-block))
  (with-fore-color *white-color*
    (let ((blocktext (if (fulltext? block)
                       (concatenate 'string (if (motorA? block) "A")
                                            (if (motorB? block) "B")
                                            " Lv"
                                            (format nil "~S" (command block)))
                       "SetPower")))
      (#_MoveTo (- (point-h (rect-centre (boundrect block))) (truncate (string-width blocktext) 2))
                (+ (point-v (rect-centre (boundrect block))) 3))
      (with-pstrs ((pstr blocktext))
        (#_DrawString pstr)))))

(defmethod draw-block-text ((block onfor-block))
  (with-fore-color *white-color*
    (let ((blocktext (concatenate 'string (if (fulltext? block) (if (motorA? block) "A"))
                                          (if (fulltext? block) (if (motorB? block) "B"))
                                          (if (fulltext? block) " ")
                                          "ONFOR")))
      (#_MoveTo (- (point-h (rect-centre (boundrect block))) 4 (truncate (string-width blocktext) 2))
                (+ (point-v (rect-centre (boundrect block))) 3))
      (with-pstrs ((pstr blocktext))
        (#_DrawString pstr)))))

(defmethod draw-block-text ((block beep-block))
  (with-fore-color *white-color*
    (let ((blocktext "BEEP"))
      (#_MoveTo (- (point-h (rect-centre (boundrect block))) (truncate (string-width blocktext) 2))
                (+ (point-v (rect-centre (boundrect block))) 3))
      (with-pstrs ((pstr blocktext))
        (#_DrawString pstr)))))

(defmethod draw-block-text ((block note-block))
  (with-fore-color *white-color*
    (let ((blocktext (if (fulltext? block)
                       (concatenate 'string (cond ((= (octave block) 1) "High ")
                                                  ((= (octave block) 2) "Mid. ")
                                                  ((= (octave block) 3) "Low "))
                                            (cond ((= (pitch block) 1) "A")
                                                  ((= (pitch block) 2) "B")
                                                  ((= (pitch block) 3) "C")
                                                  ((= (pitch block) 4) "D")
                                                  ((= (pitch block) 5) "E")
                                                  ((= (pitch block) 6) "F")
                                                  ((= (pitch block) 7) "G"))
                                            (cond ((= (modifier block) 1) "#")
                                                  ((= (modifier block) 2) "")
                                                  ((= (modifier block) 3) "b")))
                       "NOTE")))
      (#_MoveTo (- (point-h (rect-centre (boundrect block))) 4 (truncate (string-width blocktext) 2))
                (+ (point-v (rect-centre (boundrect block))) 3))
      (with-pstrs ((pstr blocktext))
        (#_DrawString pstr)))))

(defmethod draw-block-text ((block wait-block))
  (with-fore-color *white-color*
    (let ((blocktext "WAIT"))
      (#_MoveTo (- (point-h (rect-centre (boundrect block))) 4 (truncate (string-width blocktext) 2))
                (+ (point-v (rect-centre (boundrect block))) 3))
      (with-pstrs ((pstr blocktext))
        (#_DrawString pstr)))))

(defmethod draw-block-text ((block zerotimer-block))
  (with-fore-color *white-color*
    (let ((blocktext "Reset"))
      (#_MoveTo (- (point-h (rect-centre (boundrect block))) (truncate (string-width blocktext) 2))
                (- (point-v (rect-centre (boundrect block))) 2))
      (with-pstrs ((pstr blocktext))
        (#_DrawString pstr)))
    (let ((blocktext "Timer"))
      (#_MoveTo (- (point-h (rect-centre (boundrect block))) (truncate (string-width blocktext) 2))
                (+ (point-v (rect-centre (boundrect block))) 9))
      (with-pstrs ((pstr blocktext))
        (#_DrawString pstr)))))

(defmethod draw-block-text ((block timer-block))
  (with-fore-color *white-color*
    (let ((blocktext "Timer"))
      (#_MoveTo (+ (left (boundrect block)) 10)
                (+ (point-v (rect-centre (boundrect block))) 4))
      (with-pstrs ((pstr blocktext))
        (#_DrawString pstr)))))

(defmethod draw-block-text ((block dummy-block))
  (with-fore-color *white-color*
    (let ((blocktext "-----"))
      (#_MoveTo (- (point-h (rect-centre (boundrect block))) (truncate (string-width blocktext) 2))
                (+ (point-v (rect-centre (boundrect block))) 3))
      (with-pstrs ((pstr blocktext))
        (#_DrawString pstr)))))

(defmethod draw-block-text ((block when-block))
  (with-fore-color *white-color*
    (let ((blocktext "WHEN"))
      (#_MoveTo (+ (left (boundrect block)) 17)
                (+ (point-v (rect-centre (boundrect block))) 3))
      (with-pstrs ((pstr blocktext))
        (#_DrawString pstr)))))

(defmethod draw-block-text ((block whenoff-block))
  (with-fore-color *white-color*
    (let ((blocktext "WHEN"))
      (#_MoveTo (- (point-h (rect-centre (boundrect block))) (truncate (string-width blocktext) 2))
                (- (point-v (rect-centre (boundrect block))) 2))
      (with-pstrs ((pstr blocktext))
        (#_DrawString pstr)))
    (let ((blocktext "OFF"))
      (#_MoveTo (- (point-h (rect-centre (boundrect block))) (truncate (string-width blocktext) 2))
                (+ (point-v (rect-centre (boundrect block))) 9))
      (with-pstrs ((pstr blocktext))
        (#_DrawString pstr)))))

(defmethod draw-block-text ((block if-block))
  (with-fore-color *black-color*
    (let ((blocktext "IF ->"))
      (#_MoveTo (- (right (boundrect block)) 37 (string-width blocktext))
                (- (point-v (rect-centre (boundrect block))) 12))
      (with-pstrs ((pstr blocktext))
        (#_DrawString pstr)))
    (let ((blocktext "THEN ->"))
      (#_MoveTo (- (right (boundrect block)) 37 (string-width blocktext))
                (+ (point-v (rect-centre (boundrect block))) 19))
      (with-pstrs ((pstr blocktext))
        (#_DrawString pstr)))))

(defmethod draw-block-text ((block ifelse-block))
  (with-fore-color *black-color*
    (if (not (fulltext? block))
      (progn (let ((blocktext "IF ->"))
               (#_MoveTo (- (right (boundrect block)) 37 (string-width blocktext))
                         (- (point-v (rect-centre (boundrect block))) 12))
               (with-pstrs ((pstr blocktext))
                 (#_DrawString pstr)))
             (let ((blocktext "THEN ->"))
               (#_MoveTo (- (right (boundrect block)) 37 (string-width blocktext))
                         (+ (point-v (rect-centre (boundrect block))) 14))
               (with-pstrs ((pstr blocktext))
                 (#_DrawString pstr)))
             (let ((blocktext "ELSE --->"))
               (#_MoveTo (- (right (boundrect block)) 37 (string-width blocktext))
                         (+ (point-v (rect-centre (boundrect block))) 24))
               (with-pstrs ((pstr blocktext))
                 (#_DrawString pstr))))
      (progn (let ((blocktext "IF ->"))
               (#_MoveTo (- (right (boundrect block)) 72 37 (string-width blocktext))
                         (- (point-v (rect-centre (boundrect block))) 20))
               (with-pstrs ((pstr blocktext))
                 (#_DrawString pstr)))
             (let ((blocktext "THEN"))
               (#_MoveTo (- (right (boundrect block)) 91)
                         (+ (point-v (rect-centre (boundrect block))) 7))
               (with-pstrs ((pstr blocktext))
                 (#_DrawString pstr)))
             (let ((blocktext "ELSE"))
               (#_MoveTo (- (right (boundrect block)) 35)
                         (+ (point-v (rect-centre (boundrect block))) 7))
               (with-pstrs ((pstr blocktext))
                 (#_DrawString pstr)))))))

(defmethod draw-block-text ((block repeat-block))
  (with-fore-color *black-color*
    (let ((blocktext "REPEAT"))
      (#_MoveTo (+ (left (boundrect block)) 12)
                (+ (point-v (rect-centre (boundrect block))) 3))
      (with-pstrs ((pstr blocktext))
        (#_DrawString pstr)))))

(defmethod draw-block-text ((block loop-block))
  (with-fore-color *black-color*
    (let ((blocktext "LOOP"))
      (#_MoveTo (+ (left (boundrect block)) 19)
                (+ (point-v (rect-centre (boundrect block))) 3))
      (with-pstrs ((pstr blocktext))
        (#_DrawString pstr)))))

(defmethod draw-block-text ((block waituntil-block))
  (with-fore-color *black-color*
    (let ((blocktext "WaitUntil"))
      (#_MoveTo (+ (left (boundrect block)) 8)
                (+ (point-v (rect-centre (boundrect block))) 3))
      (with-pstrs ((pstr blocktext))
        (#_DrawString pstr)))))

(defmethod draw-block-text ((block switch-block))
  (with-fore-color *white-color*
    (let ((blocktext (concatenate 'string "Switch" (if (fulltext? block) (if (= (switch block) 1) "A" "B")))))
      (#_MoveTo (- (point-h (rect-centre (boundrect block))) (truncate (string-width blocktext) 2))
                (+ (point-v (rect-centre (boundrect block))) 3))
      (with-pstrs ((pstr blocktext))
        (#_DrawString pstr)))))

(defmethod draw-block-text ((block sensor-block))
  (with-fore-color *white-color*
    (let ((blocktext (concatenate 'string "Sensor" (if (fulltext? block) (if (= (sensor block) 1) "A" "B")))))
      (#_MoveTo (+ (left (boundrect block)) 10)
                (+ (point-v (rect-centre (boundrect block))) 4))
      (with-pstrs ((pstr blocktext))
        (#_DrawString pstr)))))

(defmethod draw-block-text ((block clearir-block))
  (with-fore-color *white-color*
    (let ((blocktext "Clear IR"))
      (#_MoveTo (- (point-h (rect-centre (boundrect block))) (truncate (string-width blocktext) 2))
                (+ (point-v (rect-centre (boundrect block))) 3))
      (with-pstrs ((pstr blocktext))
        (#_DrawString pstr)))))

(defmethod draw-block-text ((block newir?-block))
  (with-fore-color *white-color*
    (let ((blocktext "NewIR?"))
      (#_MoveTo (- (point-h (rect-centre (boundrect block))) (truncate (string-width blocktext) 2))
                (+ (point-v (rect-centre (boundrect block))) 3))
      (with-pstrs ((pstr blocktext))
        (#_DrawString pstr)))))

(defmethod draw-block-text ((block ir-block))
  (with-fore-color *white-color*
    (let ((blocktext "IR"))
      (#_MoveTo (+ (left (boundrect block)) 10)
                (+ (point-v (rect-centre (boundrect block))) 4))
      (with-pstrs ((pstr blocktext))
        (#_DrawString pstr)))))

(defmethod draw-block-text ((block sendir-block))
  (with-fore-color *white-color*
    (let ((blocktext "Send IR"))
      (#_MoveTo (- (point-h (rect-centre (boundrect block))) 4 (truncate (string-width blocktext) 2))
                (+ (point-v (rect-centre (boundrect block))) 3))
      (with-pstrs ((pstr blocktext))
        (#_DrawString pstr)))))

(defmethod draw-block-text ((block display-block))
  (with-fore-color *white-color*
    (let ((blocktext "Display"))
      (#_MoveTo (- (point-h (rect-centre (boundrect block))) 4 (truncate (string-width blocktext) 2))
                (+ (point-v (rect-centre (boundrect block))) 3))
      (with-pstrs ((pstr blocktext))
        (#_DrawString pstr)))))

(defmethod draw-block-text ((block number-block))
  (with-fore-color *white-color*
    (let ((blocktext (if (fulltext? block)
                       (cond ((= (genmethod block) 1) (format nil "~S" (value block)))
                             ((= (genmethod block) 2) "Random"))
                       "NUMBER")))
      (#_MoveTo (- (point-h (rect-centre (boundrect block))) -2 (truncate (string-width blocktext) 2))
                (+ (point-v (rect-centre (boundrect block))) 4))
      (with-pstrs ((pstr blocktext))
        (#_DrawString pstr)))))

(defmethod draw-block-text ((block boolean-block))
  (with-fore-color *white-color*
    (let ((blocktext (if (fulltext? block)
                       (cond ((= (comparator block) 1) "<")
                             ((= (comparator block) 2) "=")
                             ((= (comparator block) 3) ">"))
                       "?")))
      (#_MoveTo (- (point-h (rect-centre (boundrect block))) (truncate (string-width blocktext) 2))
                (+ (point-v (rect-centre (boundrect block))) 3))
      (with-pstrs ((pstr blocktext))
        (#_DrawString pstr)))))

(defmethod draw-block-text ((block and-block))
  (with-fore-color *white-color*
    (#_MoveTo (- (point-h (rect-centre (boundrect block))) (truncate (string-width "A") 2))
              (- (point-v (rect-centre (boundrect block))) 5))
    (with-pstrs ((pstr "A")) (#_DrawString pstr))
    (#_MoveTo (- (point-h (rect-centre (boundrect block))) (truncate (string-width "N") 2) -1)
              (+ (point-v (rect-centre (boundrect block))) 3))
    (with-pstrs ((pstr "N")) (#_DrawString pstr))
    (#_MoveTo (- (point-h (rect-centre (boundrect block))) (truncate (string-width "D") 2) -1)
              (+ (point-v (rect-centre (boundrect block))) 11))
    (with-pstrs ((pstr "D")) (#_DrawString pstr))))

(defmethod draw-block-text ((block or-block))
  (with-fore-color *white-color*
    (#_MoveTo (- (point-h (rect-centre (boundrect block))) (truncate (string-width "O") 2) -1)
              (- (point-v (rect-centre (boundrect block))) 1))
    (with-pstrs ((pstr "O")) (#_DrawString pstr))
    (#_MoveTo (- (point-h (rect-centre (boundrect block))) (truncate (string-width "R") 2) -1)
              (+ (point-v (rect-centre (boundrect block))) 8))
    (with-pstrs ((pstr "R")) (#_DrawString pstr))))

(defmethod draw-block-text ((block cleardata-block))
  (with-fore-color *white-color*
    (let ((blocktext "Clear"))
      (#_MoveTo (- (point-h (rect-centre (boundrect block))) (truncate (string-width blocktext) 2))
                (- (point-v (rect-centre (boundrect block))) 2))
      (with-pstrs ((pstr blocktext))
        (#_DrawString pstr)))
    (let ((blocktext "Data"))
      (#_MoveTo (- (point-h (rect-centre (boundrect block))) (truncate (string-width blocktext) 2))
                (+ (point-v (rect-centre (boundrect block))) 9))
      (with-pstrs ((pstr blocktext))
        (#_DrawString pstr)))))

(defmethod draw-block-text ((block record-block))
  (with-fore-color *white-color*
    (let ((blocktext "Record"))
      (#_MoveTo (- (point-h (rect-centre (boundrect block))) 4 (truncate (string-width blocktext) 2))
                (+ (point-v (rect-centre (boundrect block))) 3))
      (with-pstrs ((pstr blocktext))
        (#_DrawString pstr)))))

(defmethod draw-block-text ((block resetdp-block))
  (with-fore-color *white-color*
    (let ((blocktext "ResetDP"))
      (#_MoveTo (- (point-h (rect-centre (boundrect block))) (truncate (string-width blocktext) 2))
                (+ (point-v (rect-centre (boundrect block))) 3))
      (with-pstrs ((pstr blocktext))
        (#_DrawString pstr)))))

(defmethod draw-block-text ((block defproc-block))
  (with-fore-color *white-color*
    (let ((blocktext (proctitle block)))
      (#_MoveTo (- (point-h (rect-centre (boundrect block))) (truncate (string-width blocktext) 2))
                (+ (point-v (rect-centre (boundrect block))) 3))
      (with-pstrs ((pstr blocktext))
        (#_DrawString pstr)))))

(defmethod draw-block-text ((block stop-block))
  (with-fore-color *white-color*
    (let ((blocktext "STOP"))
      (#_MoveTo (- (point-h (rect-centre (boundrect block))) (truncate (string-width blocktext) 2))
                (+ (point-v (rect-centre (boundrect block))) 4))
      (with-pstrs ((pstr blocktext))
        (#_DrawString pstr)))))

(defmethod draw-block-text ((block callproc-block))
  (with-fore-color *white-color*
    (let ((blocktext (proctitle block)))
      (#_MoveTo (- (point-h (rect-centre (boundrect block))) (truncate (string-width blocktext) 2))
                (+ (point-v (rect-centre (boundrect block))) 3))
      (with-pstrs ((pstr blocktext))
        (#_DrawString pstr)))))

;--------------- BLOCK DRAG OPS ------------------------------------

(defmethod execute-drag ((view wrkarea-view) blocklist from)
  (if (< (length blocklist) 6)
    (let* ((palette-drag? (< (point-h from) 0))
           (secured?      (not palette-drag?))
           (oldpos        from)
           (curpos        from)
           (tolerances (list 0 0 0 0)))
      (rlet ((grouprect :rect :topleft  from
                              :botright from))
        (loop for block in blocklist do
              (union-rects grouprect (boundrect block) grouprect))
        (setf (car    tolerances) (- (point-h from) (left grouprect)))
        (setf (cadr   tolerances) (- (right  grouprect) (point-h from)))
        (setf (caddr  tolerances) (- (point-v from) (top  grouprect)))
        (setf (cadddr tolerances) (- (bottom grouprect) (point-v from)))
        (without-interrupts
         (loop while (mouse-down-p) do
               (setf curpos (limit-point-to-rect (view-mouse-position view)
                                                 tolerances
                                                 (if secured? (localrect view)
                                                              (hybridrect (view-container view)))))
               (unless (= curpos oldpos)
                 (execute-double-buffer (view-container view) (not secured?) blocklist (subtract-points curpos oldpos))
                 (setf oldpos curpos)
                 (if (not secured?)
                   (if (>= (- (point-h curpos) (car tolerances)) 0)
                     (setf secured? t))))))
        (if palette-drag? secured? (subtract-points from curpos))))
    (let ((movinggroup (new-region)))
      (loop for block in blocklist do
            (union-regions movinggroup (boundrgn block) movinggroup))
      (let ((grouprect (access-region-boundrect movinggroup)))
        (rlet ((limitrect  :rect :topleft  (make-point (- (point-h from) (left grouprect) -1)
                                                       (- (point-v from) (top  grouprect) -1))
                                 :botright (make-point (- (point-h (view-size view))
                                                          (- (right grouprect)  (point-h from)))
                                                       (- (point-v (view-size view))
                                                          (- (bottom grouprect) (point-v from)))))
               (screenrect :rect :topleft  (make-point (- 0 *screen-width*) (- 0 *screen-height*))
                                 :botright (make-point      *screen-width*       *screen-height*)))
          (LtG limitrect view)
          (let ((offset (#_DragGrayRgn movinggroup from limitrect screenrect 0 (%null-ptr))))
            (setf offset (GtL offset view))
            (execute-double-buffer (view-container view) nil blocklist offset)
            (dispose-region movinggroup)
            (inverse offset)))))))

;---------------

(defun limit-point-to-rect (point tolerances rect)
  (if (< (- (point-h point) (car tolerances) 1) (left rect))
    (setf point (make-point (+ (left rect) (car tolerances) 1) (point-v point)))
    (if (> (+ (point-h point) (cadr tolerances) 1) (right rect))
      (setf point (make-point (- (right rect) (cadr tolerances) 1) (point-v point)))))
  (if (< (- (point-v point) (caddr tolerances) 1) (top rect))
    (setf point (make-point (point-h point) (+ (top rect) (caddr tolerances) 1)))
    (if (> (+ (point-v point) (cadddr tolerances) 1) (bottom rect))
      (setf point (make-point (point-h point) (- (bottom rect) (cadddr tolerances) 1)))))
  point)

(defmethod execute-double-buffer ((window logoblocks-window) use-whole-window? blocklist offset)
  (gworld-copythru (buffer3 window)
                   (buffer4 window)
                   (if use-whole-window? (windowrect window) (globalrect (wrkarea window))))

  (LtG               blocklist (wrkarea window))
  (offset-blocks     blocklist offset)
  (offset-blocks     blocklist #@( 1  2))
  (loop for block in blocklist do (draw-shadow (wrkarea window) block 4))
  (offset-blocks     blocklist #@(-1 -2))
  (loop for block in blocklist do (draw-block  (wrkarea window) block 4))
  (GtL               blocklist (wrkarea window))

  (if use-whole-window?
    (gworld-copyfrom (buffer4 window) window (windowrect window))
    (gworld-copyfrom (buffer4 window)
                     (wrkarea window)
                     (globalrect (wrkarea window))
                     (localrect  (wrkarea window)))))