;--------------- SERIAL-STREAM OPS ------------------------------------

(require 'serial-streams)

(defun open-interface-serial-stream ()
  (cond ((= *interface-serial-port* 1) (make-serial-a))
        ((= *interface-serial-port* 2) (make-serial-b))))

(defun iss-clear ()
  (cond ((= *interface-serial-port* 1) (stream-clear-input *serial-a*))
        ((= *interface-serial-port* 2) (stream-clear-input *serial-b*))))

(defun iss-send (byte)
  (cond ((= *interface-serial-port* 1) (stream-tyo *serial-a* (code-char (ldb (byte 8 0) byte))))
        ((= *interface-serial-port* 2) (stream-tyo *serial-b* (code-char (ldb (byte 8 0) byte))))))

(defun iss-listen ()
  (cond ((= *interface-serial-port* 1) (stream-listen *serial-a*))
        ((= *interface-serial-port* 2) (stream-listen *serial-b*))))

(defun iss-get ()
  (cond ((= *interface-serial-port* 1) (char-code (stream-tyi *serial-a*)))
        ((= *interface-serial-port* 2) (char-code (stream-tyi *serial-b*)))))

(defun close-interface-serial-stream ()
  (cond ((= *interface-serial-port* 1) (dispose-serial-a))
        ((= *interface-serial-port* 2) (dispose-serial-b))))

;--------------- CRICKET/BRICK OPS ------------------------------------

; -4 either -2 or -3 (blue only)
; -3 no interface
; -2 interface but no power
; -1 no targets
;  0 works

(defun check-crickets ()
  (let ((result 0))
    (cond ((= *active-project-mode* 1) (if (= result 0) (setf result (send-byte 141)))
                                       (if (= result 0) (setf result (send-byte 0)))
                                       (if (= result 0) (if (not (iss-listen)) (setf result -1))))
          ((= *active-project-mode* 2) (if (= result 0) (setf result (send-byte 135)))
                                       (if (= result 0) (setf result (send-byte 0)))
                                       (if (= result 0) (if (not (iss-listen)) (setf result -1)))))
    result))

(defun download (address bytelist check-at-top? check-at-end?)
  (let ((result 0))
    (if (= result 0) (if check-at-top? (setf result (check-crickets))))
    (if (= result 0) (setf result (eeprom-setptr address)))
    (if (= result 0) (setf result (eeprom-write  bytelist)))
    (if (= result 0) (if check-at-end? (setf result (check-crickets))))
    result))

(defun set-first-ir-vector (address)
  (download (cond ((= *active-project-mode* 1) #x7e0)
                  ((= *active-project-mode* 2) #xff0))
            (list (ldb (byte 8 8) address)
                  (ldb (byte 8 0) address))
            t
            nil))

(defun set-second-ir-vector (address)
  (download (cond ((= *active-project-mode* 1) #x7e2)
                  ((= *active-project-mode* 2) #xff2))
            (list (ldb (byte 8 8) address)
                  (ldb (byte 8 0) address))
            nil
            nil))

(defun write-sensor-feeds ()
  (download (cond ((= *active-project-mode* 1) #x500)
                  ((= *active-project-mode* 2) #xec4))
            (append (trbc '<sensora>)
                    (trbc '<send>)
                    (list 0)
                    (trbc '<sensorb>)
                    (trbc '<send>)
                    (list 0))
            t
            t))

(defun get-sensor-value (offset)
  (let ((result (check-crickets)))
    (if (= result 0)
      (cond ((= *active-project-mode* 1) (if (= result 0) (setf result (eeprom-setptr (+ #x500 offset))))
                                         (if (= result 0) (setf result (send-byte 140)))
                                         (if (= result 0) (progn (sleep 0.15)
                                                                 (setf result (if (iss-listen)
                                                                                (iss-get)
                                                                                -5)))))
            ((= *active-project-mode* 2) (if (= result 0) (setf result (eeprom-setptr (+ #xec4 offset))))
                                         (if (= result 0) (setf result (send-byte 134)))
                                         (if (= result 0) (progn (sleep 0.15)
                                                                 (setf result (if (iss-listen)
                                                                                (iss-get)
                                                                                -5)))))))
    result))

;---------------

(defun eeprom-setptr (address)
  (let ((result 0))
    (if (= result 0) (setf result (send-byte (cond ((= *active-project-mode* 1) 137)
                                                   ((= *active-project-mode* 2) 131)))))
    (if (= result 0) (setf result (send-byte (ldb (byte 8 8) address))))
    (if (= result 0) (setf result (send-byte (ldb (byte 8 0) address))))
    result))

(defun eeprom-write (bytelist)
  (let ((result 0))
    (if (= result 0) (setf result (send-byte (cond ((= *active-project-mode* 1) 139)
                                                   ((= *active-project-mode* 2) 133)))))
    (if (= result 0) (setf result (send-byte (ldb (byte 8 8) (length bytelist)))))
    (if (= result 0) (setf result (send-byte (ldb (byte 8 0) (length bytelist)))))
    (dotimes (i (length bytelist))
      (if (= result 0) (setf result (send-byte (nth i bytelist)))))
    result))

(defun send-byte (byte)
  (iss-clear)
  (iss-send byte)
  (sleep 0.02)
  (cond ((= *active-project-mode* 1) (if (iss-listen)
                                       (progn (iss-get)
                                              (if (iss-listen)
                                                (progn (iss-get)
                                                       0)
                                                -2))
                                       -3))
        ((= *active-project-mode* 2) (if (iss-listen)
                                       (progn (iss-get)
                                              0)
                                       -4))))