;--------------- MENU CREATION OPS ------------------------------------

(defun create-logoblocks-menubar ()
  (let ((line-item        (make-instance 'menu-item :menu-item-title  "-"))
        (about-item       (make-instance 'menu-item :menu-item-title  "About LogoBlocks 2.5"
                                                    :menu-item-action #'(lambda () (application-about-dialog *application*))))
        (new-item         (make-instance 'menu-item :menu-item-title  "New"
                                                    :menu-item-action #'(lambda () (new-file     *window*))
                                                    :command-key      #\N))
        (open-item        (make-instance 'menu-item :menu-item-title  "Open"
                                                    :menu-item-action #'(lambda () (open-file    *window*))
                                                    :command-key      #\O))
        (save-item        (make-instance 'menu-item :menu-item-title  "Save"
                                                    :menu-item-action #'(lambda () (save-file    *window*))
                                                    :command-key      #\S
                                                    :disabled         t))
        (saveas-item      (make-instance 'menu-item :menu-item-title  "Save As"
                                                    :menu-item-action #'(lambda () (save-file-as *window*))))
        (quit-item        (make-instance 'menu-item :menu-item-title  "Quit"
                                                    :menu-item-action #'(lambda () (if (save?-dialog *window*)
                                                                                     (progn (close-interface-serial-stream)
                                                                                            (quit))))
                                                    :command-key      #\Q))
        (selectall-item   (nth 7 (menu-items *edit-menu*)))
        (preferences-item (make-instance 'menu-item :menu-item-title  "Preferences"
                                                    :menu-item-action #'(lambda () (preferences-dialog *window*))
                                                    :command-key      #\P))
        (download-item    (make-instance 'menu-item :menu-item-title  "Download P1/P2"
                                                    :menu-item-action #'(lambda () (download-p1/p2 (wrkarea *window*)))
                                                    :command-key      #\D))
        (run-item         (make-instance 'menu-item :menu-item-title  "Run Selected Blocks"
                                                    :command-key      #\R
                                                    :disabled         t))
        (convert-item     (make-instance 'menu-item :menu-item-title  "Convert Project"
                                                    :command-key      #\T
                                                    :disabled         t))
        (monitor-item     (make-instance 'menu-item :menu-item-title  "Cricket Sensor Monitor"
                                                    :menu-item-action #'(lambda () (monitor-dialog))
                                                    :command-key      #\M)))
    (apply #'remove-menu-items *apple-menu* (menu-items *apple-menu*))
    (add-menu-items            *apple-menu* about-item line-item)
    (remove-menu-items *edit-menu*  (nth  0 (menu-items *edit-menu*))
                                    (nth  1 (menu-items *edit-menu*))
                                    (nth  2 (menu-items *edit-menu*))
                                    (nth  7 (menu-items *edit-menu*))
                                    (nth  8 (menu-items *edit-menu*))
                                    (nth  9 (menu-items *edit-menu*))
                                    (nth 10 (menu-items *edit-menu*))
                                    (nth 11 (menu-items *edit-menu*))
                                    (nth 12 (menu-items *edit-menu*))
                                    (nth 13 (menu-items *edit-menu*))
                                    (nth 14 (menu-items *edit-menu*))
                                    (nth 15 (menu-items *edit-menu*))
                                    (nth 16 (menu-items *edit-menu*))
                                    (nth 17 (menu-items *edit-menu*)))
    (add-menu-items    *edit-menu*         line-item
                                      selectall-item
                                           line-item
                                    preferences-item)
    (let ((file-menu     (make-instance 'menu :menu-title "File"
                                              :menu-items (list new-item open-item save-item saveas-item line-item quit-item)))
          (project-menu  (make-instance 'menu :menu-title "Project"
                                              :menu-items (list download-item run-item line-item convert-item line-item monitor-item)))
          (examples-menu (make-instance 'menu :menu-title "Examples")))
      (set-menubar (list file-menu
                         *edit-menu*
                         project-menu
                         examples-menu))
      (setf *save-item*     save-item)
      (setf *examples-menu* examples-menu))))

;--------------- MENU MODIFICATION OPS ------------------------------------

(defun set-flag ()
  (unless *save-flag*
    (menu-item-enable *save-item*)
    (setf *save-flag* t)))

(defun clear-flag ()
  (unless (not *save-flag*)
    (menu-item-disable *save-item*)
    (setf *save-flag* nil)))

(defun rebuild-examples-menu ()
  (apply #'remove-menu-items *examples-menu* (menu-items *examples-menu*))
  (let ((examples-path (cond ((= *default-project-mode* 1) "CCL:Red Dot Examples;")
                             ((= *default-project-mode* 2) "CCL:Blue Dot Examples;")
                             ((= *default-project-mode* 3) "CCL:LEGO RCX Examples;")))
        (example-names (cond ((= *default-project-mode* 1) '("Forward & Stop" "Back & Forth" "Find Light" "Send IR" "Receive IR" "Collect Data" "Blinky"))
                             ((= *default-project-mode* 2) '("Forward & Stop" "Back & Forth" "Find Light" "Send IR" "Receive IR" "Collect Data" "Background Process" "Blinky"))
                             ((= *default-project-mode* 3) '()))))
    (loop for x in (first-n-integers (length example-names)) do
          (let* ((example-file-name (concatenate 'string "Example " (format nil "~S" (+ x 1)) " (" (nth x example-names) ")"))
                 (example-file-path (concatenate 'string examples-path example-file-name)))
            (add-menu-items *examples-menu* (make-instance 'menu-item :menu-item-title  example-file-name
                                                                      :menu-item-action #'(lambda () (open-file *window* example-file-path))
                                                                      :command-key      (if (< x 10) (character (format nil "~S" (+ x 1))))))))))

;--------------- WINDOW CLASSES/SSMs ------------------------------------

(defclass logoblocks-window (window)
  ((windowrect                                   :accessor windowrect)
   (hybridrect                                   :accessor hybridrect)
   (buttons                  :initarg :buttons   :reader   buttons)
   (palette                  :initarg :palette   :reader   palette)
   (wrkarea                  :initarg :wrkarea   :reader   wrkarea)
   (scrollbar                :initarg :scrollbar :reader   scrollbar)
   (buffer3                                      :accessor buffer3)     ; static composite
   (buffer4                                      :accessor buffer4)))   ; motion composite

(defmethod view-key-event-handler ((window logoblocks-window) char)
  (keyboard-handler window char))

(defmethod select-all ((window logoblocks-window))
  (set-flag)
  (select-all-blocks (wrkarea window)))

;--------------- VIEW CLASSES/SSMs ------------------------------------

(defclass logoblocks-view (view)
  ((localrect                                    :accessor localrect)
   (globalrect                                   :accessor globalrect)))

;---------------

(defclass scrollbar-view (scroll-bar-dialog-item
                            logoblocks-view)
  ((data                                         :accessor data)))

(defclass original-view (logoblocks-view)
  ())

(defmethod view-draw-contents ((view original-view))
  (update-handler view))

(defmethod view-click-event-handler ((view original-view) where)
  (mouse-handler view where))

;---------------

(defclass buttons-view (original-view)
  ((controls                                     :accessor controls)
   (switch                                       :accessor switch)))

(defclass working-view (original-view)
  ((offsetrect                                   :accessor offsetrect)
   (masterlist :initform nil                     :accessor masterlist)
   (buffer1                                      :accessor buffer1)     ; block shadows
   (buffer2                                      :accessor buffer2)))   ; block images and text

(defclass palette-view (working-view)
  ())

(defclass wrkarea-view (working-view)
  ())

;--------------- WINDOW CREATION OPS ------------------------------------

(defun create-logoblocks-window (&optional (switch     0)
                                               (blocklist1 nil)
                                               (blocklist2 nil))
  (let ((buttons   (make-instance 'buttons-view      :view-position    (cond ((= *current-window-size* 1) #@(   0 275))
                                                                             ((= *current-window-size* 2) #@(   0 395))
                                                                             ((= *current-window-size* 3) #@(   0 563)))
                                                     :view-size        #@(126 150)))
        (palette   (make-instance 'palette-view      :view-position    #@(  0   0)
                                                     :view-size        (cond ((= *current-window-size* 1) #@( 111 275))
                                                                             ((= *current-window-size* 2) #@( 111 395))
                                                                             ((= *current-window-size* 3) #@( 111 563)))))
        (wrkarea   (make-instance 'wrkarea-view      :view-position    #@(126   0)
                                                     :view-size        (cond ((= *current-window-size* 1) #@( 494 425))
                                                                             ((= *current-window-size* 2) #@( 654 545))
                                                                             ((= *current-window-size* 3) #@( 878 713)))))
        (scrollbar (make-instance 'scrollbar-view    :view-position    #@(110   0)
                                                     :view-size        (cond ((= *current-window-size* 1) #@(  16 275))
                                                                             ((= *current-window-size* 2) #@(  16 395))
                                                                             ((= *current-window-size* 3) #@(  16 563)))
                                                     :dialog-item-text ""
                                                     :direction        :vertical
                                                     :min              0
                                                     :scroll-size      1
                                                     :page-size        2
                                                     :track-thumb-p    t)))
    (setf *window* (make-instance 'logoblocks-window :view-position    #@( 10  45)
                                                     :view-size        (cond ((= *current-window-size* 1) #@( 620 425))
                                                                             ((= *current-window-size* 2) #@( 780 545))
                                                                             ((= *current-window-size* 3) #@(1004 713)))
                                                     :auto-position    (if (or (and (= *current-window-size*   1)
                                                                                    (= *screen-width*        640)
                                                                                    (= *screen-height*       480))
                                                                               (and (= *current-window-size*   2)
                                                                                    (= *screen-width*        800)
                                                                                    (= *screen-height*       600))
                                                                               (and (= *current-window-size*   3)
                                                                                    (= *screen-width*       1024)
                                                                                    (= *screen-height*       768)))
                                                                         :NoAutoCenter
                                                                         :CenterParentWindowScreen)
                                                     :window-type      :document
                                                     :close-box-p      nil
                                                     :color-p          t
                                                     :window-show      nil
                                                     :buttons          buttons
                                                     :palette          palette
                                                     :wrkarea          wrkarea
                                                     :scrollbar        scrollbar)))
  (update-window-title)
  (alloc/init-window             *window*)
  (alloc/init-view    (buttons   *window*) *window* switch)
  (alloc/init-view    (palette   *window*) *window* nil)
  (alloc/init-view    (wrkarea   *window*) *window* nil)
  (alloc/init-view    (scrollbar *window*) *window* nil)
  (reset-view         (scrollbar *window*) nil)
  (reset-view         (palette   *window*) blocklist1)
  (reset-view         (wrkarea   *window*) blocklist2)
  (draw-buffer3       (buttons   *window*) nil nil)
  (draw-buffer3       (palette   *window*) (masterlist (palette *window*)) 'draw)
  (draw-buffer3       (wrkarea   *window*) (masterlist (wrkarea *window*)) 'draw)
  (view-draw-contents (buttons   *window*))
  (view-draw-contents (palette   *window*))
  (view-draw-contents (wrkarea   *window*))
  (view-draw-contents (scrollbar *window*))
  (window-show                   *window*))

;--------------- WINDOW MODIFICATION OPS ------------------------------------

(defun update-window-title ()
  (set-window-title *window* (concatenate 'string (cond ((= *active-project-mode* 1)  "Red Dot Project  ")
                                                        ((= *active-project-mode* 2) "Blue Dot Project  ")
                                                        ((= *active-project-mode* 3) "LEGO RCX Project  "))
                                                  (if *file-path* (pathname-name *file-path*) "Untitled"))))

;--------------- WINDOW/VIEW ALLOC/INIT OPS ------------------------------------

(defmethod alloc/init-window ((window logoblocks-window))
  (setf (windowrect window) (make-record :rect :topleft  #@(0 0)
                                               :botright (view-size window)))
  (setf (hybridrect window) (make-record :rect :topleft  (subtract-points (view-position (palette window))
                                                                          (view-position (wrkarea window)))
                                               :botright (make-point (point-h (view-size (wrkarea window)))
                                                                     (point-v (view-size (palette window))))))
  (setf    (buffer3 window) (create-gworld (windowrect window) 8))
  (setf    (buffer4 window) (create-gworld (windowrect window) 8))
  (multiple-value-bind (ff ms) (font-codes '("Geneva" 9))
    (set-wptr-font-codes (buffer4 window) ff ms)))

(defmethod alloc/init-view ((view buttons-view) (parent logoblocks-window) parameter)
  (set-view-container view parent)
  (setf (localrect  view) (make-record :rect :topleft  #@(0 0)
                                             :botright (view-size view)))
  (setf (globalrect view) (LtG (duplicate-rect (localrect view)) view))
  (let ((rect1 (make-record :rect :topleft #@( 9  10) :botright #@( 84  38)))
        (rect2 (make-record :rect :topleft #@( 9  44) :botright #@( 84  72)))
        (rect3 (make-record :rect :topleft #@( 9  78) :botright #@( 84 106)))
        (rect4 (make-record :rect :topleft #@( 9 112) :botright #@(115 140)))
        (rect5 (make-record :rect :topleft #@(90  10) :botright #@(115 106))))
    (setf (controls view) (make-array 5 :initial-contents (list rect1 rect2 rect3 rect4 rect5))))
  (setf (switch     view) parameter))

(defmethod alloc/init-view ((view palette-view) (parent logoblocks-window) parameter)
  (declare (ignore parameter))
  (set-view-container view parent)
  (setf (localrect  view) (make-record :rect :topleft  #@(0 0)
                                             :botright (view-size view)))
  (setf (globalrect view) (LtG         (duplicate-rect (localrect  view)) view))
  (setf (offsetrect view) (offset-rect (duplicate-rect (globalrect view)) #@(1 2)))
  (setf         (buffer1 view) (create-gworld (localrect view) 1))
  (setf         (buffer2 view) (create-gworld (localrect view) 8))
  (do-in-gworld (buffer1 view) (with-back-color *white-color* (#_EraseRect (localrect view))))
  (do-in-gworld (buffer2 view) (with-back-color *white-color* (#_EraseRect (localrect view))))
  (multiple-value-bind (ff ms) (font-codes '("Geneva" 9))
    (set-wptr-font-codes (buffer2 view) ff ms)))

(defmethod alloc/init-view ((view wrkarea-view) (parent logoblocks-window) parameter)
  (declare (ignore parameter))
  (set-view-container view parent)
  (setf (localrect  view) (make-record :rect :topleft  #@(0 0)
                                             :botright (view-size view)))
  (setf (globalrect view) (LtG         (duplicate-rect (localrect  view)) view))
  (setf (offsetrect view) (offset-rect (duplicate-rect (globalrect view)) #@(1 2)))
  (setf         (buffer1 view) (create-gworld (localrect view) 1))
  (setf         (buffer2 view) (create-gworld (localrect view) 8))
  (do-in-gworld (buffer1 view) (with-back-color *white-color* (#_EraseRect (localrect view))))
  (do-in-gworld (buffer2 view) (with-back-color *white-color* (#_EraseRect (localrect view))))
  (multiple-value-bind (ff ms) (font-codes '("Geneva" 9))
    (set-wptr-font-codes (buffer2 view) ff ms)))

(defmethod alloc/init-view ((view scrollbar-view) (parent logoblocks-window) parameter)
  (declare (ignore parameter))
  (set-view-container view parent)
  (set-dialog-item-action-function view #'(lambda (item) item (draw-buffer3       (palette (view-container view)) (masterlist (palette (view-container view))) 'erase)
                                                              (scrollset-blocks   (palette (view-container view)))
                                                              (draw-buffer3       (palette (view-container view)) (masterlist (palette (view-container view))) 'draw)
                                                              (view-draw-contents (palette (view-container view)))))
  (setf (localrect  view) (make-record :rect :topleft  #@(0 0)
                                             :botright (view-size view)))
  (setf (globalrect view) (LtG (duplicate-rect (localrect view)) view)))

;--------------- VIEW RESET OPS ------------------------------------

(defmethod reset-view ((view palette-view) blocklist)
  (let ((switch      (switch (buttons (view-container view))))
        (specieslist nil)
        (curpos      10))
    (if blocklist
      (progn (setf (masterlist view) blocklist)
             (offset-blocks (masterlist view) (make-point 0 (- 10 (top (boundrect (car (masterlist view))))))))
      (progn (cond ((= switch 0) (setf specieslist (list 'on/off 'direction 'power 'onfor 'separator 'beep 'note 'separator 'wait 'zerotimer 'timer 'separator 'dummy 'number)))
                   ((= switch 1) (setf specieslist (list 'if 'ifelse 'repeat 'loop 'waituntil 'separator 'number))
                                 (if (= *active-project-mode* 2)
                                   (progn (setf specieslist (append (list 'whenoff 'separator) specieslist))
                                          (if (not (index-when-block (wrkarea (view-container view))))
                                            (setf specieslist (cons 'when specieslist))))))
                   ((= switch 2) (setf specieslist (append (list 'switch 'sensor 'separator)
                                                           (cond ((= *active-project-mode* 1) (list 'clearir))
                                                                 ((= *active-project-mode* 2) (list 'newir?)))
                                                           (list 'ir 'sendir 'separator 'display 'separator 'number 'boolean 'and 'or 'separator 'cleardata 'record 'resetdp))))
                   ((= switch 3) (setf specieslist (list (list 'defproc   (list 0 ""))
                                                         (list 'separator nil)
                                                         (list 'stop      nil)
                                                         (list 'separator nil)
                                                         (list 'callproc  (list 0 ""))))
                                 (if (not (index-defproc-block (wrkarea (view-container view)) 2))
                                   (setf specieslist (append (list (list 'defproc (list 2 "P2"))) specieslist)))
                                 (loop for block in (masterlist (wrkarea (view-container view))) do
                                       (if (equal (species block) 'defproc)
                                         (if (> (procnum block) 2)
                                           (setf specieslist (append specieslist (list (list 'callproc (list (procnum   block)
                                                                                                             (proctitle block))))))))))
                   ((= switch 4) (setf specieslist (list))))
             (loop for x in (first-n-integers (length specieslist)) do
                   (let ((species  (nth x specieslist))
                         (procdata nil))
                     (if (= switch 3) (progn (setf procdata (cadr species)) (setf species (car species))))
                     (create-block view species nil nil (make-point (block-h-offset switch (block-family species nil)) curpos) procdata)
                     (setf curpos (+ curpos (rect-height (boundrect (car (last (masterlist view))))) 10))))))
    (let ((x        0)
          (proceed? t))
      (loop while proceed? do
            (set-block-visibilities view)
            (if (= (count-lower-invisibles view) 0)
              (setf proceed? nil)
              (let ((offset (- 0 (+ (rect-height (boundrect (nth x (masterlist view)))) 10))))
                (offset-blocks (masterlist view) (make-point 0 offset))
                (setf x (+ x 1)))))
      (set-scroll-bar-max     (scrollbar (view-container view)) x)
      (set-scroll-bar-setting (scrollbar (view-container view)) (nth switch (data (scrollbar (view-container view)))))
      (scrollset-blocks view))))

(defmethod reset-view ((view wrkarea-view) blocklist)
  (if (not blocklist)
    (create-block view 'defproc t nil #@(90 30) (list 1 "P1"))
    (setf (masterlist view) blocklist)))

(defmethod reset-view ((view scrollbar-view) blocklist)
  (declare (ignore blocklist))
  (setf (data view) (list 0 0 0 0 0)))

;--------------- UPDATE HANDLERS ------------------------------------

(defmethod update-handler ((view original-view))
  (gworld-copyfrom (buffer3 (view-container view))
                   view
                   (globalrect view)
                   (localrect  view)))

;--------------- KEYBOARD HANDLERS ------------------------------------

(defmethod keyboard-handler ((window logoblocks-window) char)
  (cond ((equal char #\Delete)    (set-flag)
                                  (delete-selected-blocks (wrkarea window)))
        ((equal char #\UpArrow)   (if (> (scroll-bar-setting (scrollbar window))
                                         (scroll-bar-min     (scrollbar window)))
                                    (progn (set-scroll-bar-setting (scrollbar window) (- (scroll-bar-setting (scrollbar window)) 1))
                                           (draw-buffer3           (palette   window) (masterlist (palette window)) 'erase)
                                           (scrollset-blocks       (palette   window))
                                           (draw-buffer3           (palette   window) (masterlist (palette window)) 'draw)
                                           (view-draw-contents     (palette   window))
                                           (view-draw-contents     (scrollbar window)))))
        ((equal char #\DownArrow) (if (< (scroll-bar-setting (scrollbar window))
                                         (scroll-bar-max     (scrollbar window)))
                                    (progn (set-scroll-bar-setting (scrollbar window) (+ (scroll-bar-setting (scrollbar window)) 1))
                                           (draw-buffer3           (palette   window) (masterlist (palette window)) 'erase)
                                           (scrollset-blocks       (palette   window))
                                           (draw-buffer3           (palette   window) (masterlist (palette window)) 'draw)
                                           (view-draw-contents     (palette   window))
                                           (view-draw-contents     (scrollbar window)))))))

;--------------- MOUSE HANDLERS ------------------------------------

(defmethod mouse-handler ((view buttons-view) where)
  (let ((ignore? t))
    (loop for x in (first-n-integers 4) do
          (if (point-in-rect? where (aref (controls view) x))
            (unless (equal (switch view) x)
              (setf (nth (switch view) (data (scrollbar (view-container view)))) (scroll-bar-setting (scrollbar (view-container view))))
              (setf (switch view) x)
              (setf ignore? nil))))
    (unless ignore? (draw-buffer3       view nil nil)
                    (view-draw-contents view)
                    (play-sound-sync    'click)
                    (delete-all-blocks  (palette (view-container view)))
                    (rebuild-all-blocks (palette (view-container view))))))

(defmethod mouse-handler ((view working-view) where)
  (let ((block (acquire-target view where)))
    (if block (block-click-handler view block where)
              (empty-click-handler view       where))))

;---------------

(defmethod block-click-handler ((view palette-view) block where)
  (set-flag)
  (deselect-all-blocks (wrkarea (view-container view)))
  (if (or      (equal (species block) 'when)
          (and (equal (species block) 'defproc)
               (=     (procnum block) 2)))
    (draw-buffer3 view (list block) 'erase))
  (gworld-copyto (view-container view)
                 (buffer3 (view-container view))
                 (globalrect (scrollbar (view-container view))))
  (let ((block1 (create-block (wrkarea (view-container view))
                              (species block) t t
                              (LtGtL (add-points (topleft (boundrect block))
                                                 #@(4 4))
                                     view (wrkarea (view-container view)))
                              (if (or (equal (species block) 'defproc)
                                      (equal (species block) 'callproc))
                                (list (procnum   block)
                                      (proctitle block))
                                nil)))
        (block2 (if (match? (species block) (list 'onfor 'note 'wait 'repeat 'sendir 'display 'record))
                  (create-block (wrkarea (view-container view))
                                (if (or (equal (species block) 'display)
                                        (equal (species block) 'record))
                                  'sensor
                                  'number) t t
                                (LtGtL (add-points (add-points (topleft (boundrect block))
                                                               #@(4 4))
                                                   (dock-offset (family block) 'number1))
                                       view (wrkarea (view-container view)))
                                (cond ((equal (species block) 'onfor)   10)
                                      ((equal (species block) 'note)    10)
                                      ((equal (species block) 'wait)    10)
                                      ((equal (species block) 'repeat)   5)
                                      ((equal (species block) 'sendir)   0)
                                      ((equal (species block) 'display) nil)
                                      ((equal (species block) 'record)  nil)))
                  nil)))
    (drag-blocks view (if block2 (list block1 block2) (list block1))
                 (LtGtL where view (wrkarea (view-container view))))))

(defmethod block-click-handler ((view wrkarea-view) block where)
  (set-flag)
  (if (just-a-click? view where)
    (if (shift-key-p)
      (invert-block-selections view (list block))
      (if (double-click-p)
        (flipside-dialog view block)
        (progn (deselect-all-blocks view)
               (select-block view block))))
    (progn (if (shift-key-p)
             (invert-block-selections view (list block))
             (progn (if (not (selected? block))
                      (deselect-all-blocks view))
                    (select-block view block t)))
           (drag-blocks view (get-selected-blocks view) where))))

(defmethod empty-click-handler ((view palette-view) where)
  (declare (ignore view where)))

(defmethod empty-click-handler ((view wrkarea-view) where)
  (set-flag)
  (unless (shift-key-p) (deselect-all-blocks view))
  (if (not (just-a-click? view where))
    (rlet ((dragrect :rect :topleft  where
                           :botright where))
      (anchor-based-box        view where dragrect (list 'tplf where) t t)
      (invert-block-selections view (blocks-in-rect view dragrect)))))

;---------------

(defmethod just-a-click? ((view working-view) where)
  (let ((flag  t)
        (loop? t)
        (mark  (#_TickCount)))
    (loop while (and loop? (mouse-down-p)) do
          (if (or (not (equal (view-mouse-position view) where))
                  (> (- (#_TickCount) mark) 40))
            (progn (setf flag  nil)
                   (setf loop? nil))))
    flag))

(defmethod anchor-based-box ((view wrkarea-view) where dragrect anchor h? v?)
  (unless (and (not h?) (not v?))
    (without-interrupts
     (let ((oldmouse   where)
           (oldpenptrn (pen-pattern view))
           (oldpenmode (pen-mode    view)))
       (set-pen-pattern view *gray-pattern*)
       (set-pen-mode    view :patXor)
       (with-focused-view view
         (#_FrameRect dragrect)
         (loop while (mouse-down-p) do
               (let ((newmouse (view-mouse-position view)))
                 (let ((quad (which-quadrant? anchor newmouse)))
                   (unless (cond ((and h? v?) (equal oldmouse
                                                     newmouse))
                                 (h?          (equal (point-h oldmouse)
                                                     (point-h newmouse)))
                                 (v?          (equal (point-v oldmouse)
                                                     (point-v newmouse))))
                     (#_FrameRect dragrect)
                     (if (update-drag-rect              anchor quad dragrect newmouse h? v?)
                       (setf anchor (reassert-drag-rect anchor quad dragrect newmouse h? v?)))
                     (#_FrameRect dragrect))
                   (setf oldmouse newmouse))))
         (#_FrameRect dragrect))
       (set-pen-pattern view oldpenptrn)
       (set-pen-mode    view oldpenmode)
       dragrect))))

;---------------

(defun update-drag-rect (anchor quad dragrect newmouse h? v?)
  (let ((diffquad? nil))
    (case (car anchor)
      (tplf (if (equal quad 'btrt) 
              (progn (if h? (rset dragrect :rect.right  (point-h newmouse)))
                     (if v? (rset dragrect :rect.bottom (point-v newmouse))))
              (setf diffquad? t)))
      (tprt (if (equal quad 'btlf)      
              (progn (if h? (rset dragrect :rect.left   (point-h newmouse)))
                     (if v? (rset dragrect :rect.bottom (point-v newmouse))))
              (setf diffquad? t)))
      (btlf (if (equal quad 'tprt)
              (progn (if h? (rset dragrect :rect.right  (point-h newmouse)))
                     (if v? (rset dragrect :rect.top    (point-v newmouse))))
              (setf diffquad? t)))
      (btrt (if (equal quad 'tplf)
              (progn (if h? (rset dragrect :rect.left   (point-h newmouse)))
                     (if v? (rset dragrect :rect.top    (point-v newmouse))))
              (setf diffquad? t))))
    diffquad?))

(defun reassert-drag-rect (anchor quad dragrect newmouse h? v?)
  (let ((newanchor anchor))
    (case quad
      (tplf (progn (if h? (progn (rset dragrect :rect.left   (point-h newmouse))
                                 (rset dragrect :rect.right  (point-h (cadr anchor)))))
                   (if v? (progn (rset dragrect :rect.top    (point-v newmouse))
                                 (rset dragrect :rect.bottom (point-v (cadr anchor)))))
                   (setf newanchor (list 'btrt (botright dragrect)))))
      (tprt (progn (if h? (progn (rset dragrect :rect.left   (point-h (cadr anchor)))
                                 (rset dragrect :rect.right  (point-h newmouse))))
                   (if v? (progn (rset dragrect :rect.top    (point-v newmouse))
                                 (rset dragrect :rect.bottom (point-v (cadr anchor)))))
                   (setf newanchor (list 'btlf (botleft  dragrect)))))
      (btlf (progn (if h? (progn (rset dragrect :rect.left   (point-h newmouse))
                                 (rset dragrect :rect.right  (point-h (cadr anchor)))))
                   (if v? (progn (rset dragrect :rect.top    (point-v (cadr anchor)))
                                 (rset dragrect :rect.bottom (point-v newmouse))))
                   (setf newanchor (list 'tprt (topright dragrect)))))
      (btrt (progn (if h? (progn (rset dragrect :rect.left   (point-h (cadr anchor)))
                                 (rset dragrect :rect.right  (point-h newmouse))))
                   (if v? (progn (rset dragrect :rect.top    (point-v (cadr anchor)))
                                 (rset dragrect :rect.bottom (point-v newmouse))))
                   (setf newanchor (list 'tplf (topleft  dragrect))))))
    newanchor))

(defun which-quadrant? (anchor where)
  (let ((anchorpoint (cadr anchor)))
    (cond ((point<= anchorpoint where)                'btrt)
          ((point<= where anchorpoint)                'tplf)
          ((<= (point-h where) (point-h anchorpoint)) 'btlf)
          ((<= (point-v where) (point-v anchorpoint)) 'tprt))))