;--------------- GWORLD OPS ------------------------------------

(defun create-gworld (rect depth)
  (rlet ((result :pointer))
    (#_NewGWorld result depth rect (%null-ptr) (%null-ptr) 0)
    (%get-ptr result)))

(defun dispose-gworld (gworld)
  (#_DisposeGWorld gworld))

(defmacro do-in-gworld (gworld &rest forms)
  `(rlet ((port :pointer)
          (gdh  :pointer))
     (#_LockPixels (#_GetGWorldPixMap ,gworld))
     (#_GetGWorld port gdh)
     (#_SetGWorld ,gworld (%null-ptr))
     (unwind-protect (progn ,@forms)
       (#_SetGWorld (%get-ptr port) (%get-ptr gdh))
       (#_UnlockPixels (#_GetGWorldPixMap ,gworld)))))

(defun gworld-copyfrom (gworld view rect1 &optional (rect2 rect1) (region (%null-ptr)))
  (with-focused-view view
    (#_LockPixels   (#_GetGWorldPixMap gworld))
    (#_CopyBits     (%inc-ptr gworld 2) (%inc-ptr (wptr view) 2) rect1 rect2 #$SrcCopy region)
    (#_UnlockPixels (#_GetGWorldPixMap gworld))))

(defun gworld-copyto (view gworld rect1 &optional (rect2 rect1) (region (%null-ptr)))
  (rlet ((port :pointer)
         (gdh  :pointer))
    (#_LockPixels   (#_GetGWorldPixMap gworld))
    (#_GetGWorld port gdh)
    (#_SetGWorld gworld (%null-ptr))
    (#_CopyBits     (%inc-ptr (wptr view) 2) (%inc-ptr gworld 2) rect1 rect2 #$SrcCopy region)
    (#_SetGWorld    (%get-ptr port) (%get-ptr gdh))
    (#_UnlockPixels (#_GetGWorldPixMap gworld))))

(defun gworld-copythru (gworld1 gworld2 rect1 &optional (rect2 rect1) (region (%null-ptr)))
  (#_LockPixels   (#_GetGWorldPixMap gworld1))
  (#_LockPixels   (#_GetGWorldPixMap gworld2))
  (#_CopyBits     (%inc-ptr gworld1 2)
                  (%inc-ptr gworld2 2) rect1 rect2 #$SrcCopy region)
  (#_UnlockPixels (#_GetGWorldPixMap gworld1))
  (#_UnlockPixels (#_GetGWorldPixMap gworld2)))

(defun gworld-maskthru (gworld1 gworld2 gworld3 rect1 &optional (rect2 rect1) (rect3 rect1))
  (#_LockPixels   (#_GetGWorldPixMap gworld1))
  (#_LockPixels   (#_GetGWorldPixMap gworld2))
  (#_LockPixels   (#_GetGWorldPixMap gworld3))
  (#_CopyMask     (%inc-ptr gworld1 2)
                  (%inc-ptr gworld2 2)
                  (%inc-ptr gworld3 2) rect1 rect2 rect3)
  (#_UnlockPixels (#_GetGWorldPixMap gworld1))
  (#_UnlockPixels (#_GetGWorldPixMap gworld2))
  (#_UnlockPixels (#_GetGWorldPixMap gworld3)))

(defun access-gworld-boundrect (gworld)
  (pref gworld :grafport.portrect))

(defun create-gworld-boundrgn (gworld)
  (let ((region (new-region)))
    (#_BitMapToRegion region (%inc-ptr gworld 2))
    region))

;--------------- POINT OPS ------------------------------------

(defmethod LtG ((thing fixnum) view)
  (add-points      thing (view-position view)))

(defmethod GtL ((thing fixnum) view)
  (subtract-points thing (view-position view)))

(defun inverse (point)
  (subtract-points #@(0 0) point))

(defun distance (point1 point2)
  (sqrt (+ (* (- (point-h point1) (point-h point2)) (- (point-h point1) (point-h point2)))
           (* (- (point-v point1) (point-v point2)) (- (point-v point1) (point-v point2))))))

(defun point-in-rect? (point rect)
  (#_PtInRect point rect))

(defun point-in-region? (point region)
  (#_PtInRgn  point region))

;--------------- RECT OPS ------------------------------------

(defun top      (rect) (pref rect :rect.top))
(defun left     (rect) (pref rect :rect.left))
(defun bottom   (rect) (pref rect :rect.bottom))
(defun right    (rect) (pref rect :rect.right))
(defun topleft  (rect) (pref rect :rect.topleft))
(defun botright (rect) (pref rect :rect.botright))
(defun botleft  (rect) (make-point (left  rect) (bottom rect)))
(defun topright (rect) (make-point (right rect) (top    rect)))

(defun rect-width  (rect) (- (right  rect)
                               (left   rect)))
(defun rect-height (rect) (- (bottom rect)
                               (top    rect)))
(defun rect-centre (rect) (add-points (topleft rect)
                                        (make-point (round (rect-width  rect) 2)
                                                    (round (rect-height rect) 2))))
(defun rect-size   (rect) (subtract-points (botright rect)
                                             (topleft  rect)))

(defmethod LtG ((thing macptr) view)
  (offset-rect thing (view-position view)))

(defmethod GtL ((thing macptr) view)
  (offset-rect thing (inverse (view-position view))))

(defun union-rects (rect1 rect2 rect3)
  (#_UnionRect rect1 rect2 rect3)
  rect3)

(defun rect-in-region? (rect region)
  (#_RectInRgn rect region))

(defun offset-rect (rect offset)
  (#_OffsetRect rect (point-h offset) (point-v offset))
  rect)

(defun move-rect-to (rect position)
  (offset-rect rect (subtract-points position (topleft rect))))

(defun duplicate-rect (rect)
  (make-record :rect :topleft  (topleft  rect)
                     :botright (botright rect)))

;--------------- REGION OPS ------------------------------------

(defun new-region ()
  (#_NewRgn))

(defun dispose-region (region)
  (#_DisposeRgn region))

(defun intersect-regions (region1 region2 &optional (region3 (new-region)))
  (#_SectRgn region1 region2 region3)
  region3)

(defun union-regions (region1 region2 &optional (region3 (new-region)))
  (#_UnionRgn region1 region2 region3)
  region3)

(defun empty-region? (region)
  (#_EmptyRgn region))

(defun access-region-boundrect (region)
  (with-dereferenced-handles ((rgnptr region))
    (pref rgnptr :region.rgnbbox)))

(defun offset-region (region offset)
  (#_OffsetRgn region (point-h offset) (point-v offset))
  region)

(defun move-region-to (region position)
  (offset-region region (subtract-points position (topleft (access-region-boundrect region)))))

;--------------- PICT OPS ------------------------------------

(defun access-pict-boundrect (pict)
  (with-dereferenced-handles ((pictptr pict))
    (pref pictptr :picture.picframe)))

(defun load-picts ()
  (setf *picts* (list))
  (loop for x in (first-n-integers *number-of-picts*) do
        (setf *picts* (append *picts* (list (#_GetResource "PICT" (+ 128 x)))))))

(defun dispose-picts ()
  (loop for pict in *picts* do
        (#_ReleaseResource pict))
  (setf *picts* nil))

(defun alloc/init-pict-buffers ()
  (setf *pict-buffers* (list))
  (loop for x in (first-n-integers *number-of-picts*) do
        (setf *pict-buffers* (append *pict-buffers* (list (create-gworld (access-pict-boundrect (nth x *picts*)) (if (mask-pict? x) 1 8))))))
  (loop for x in (first-n-integers *number-of-picts*) do
        (do-in-gworld (nth x *pict-buffers*) (#_DrawPicture (nth x *picts*) (access-gworld-boundrect (nth x *pict-buffers*))))))

;---------------

(defun mask-pict? (x)
  (multiple-value-bind (a b) (truncate (+ x 1) 3)
    (or (and (<= a 24) (= b 0))
        (= x 73)
        (= x 74))))

;--------------- SOUND OPS ------------------------------------

(defun play-sound-sync (name)
  (#_SndPlay (%null-ptr) (nth (index-sound name) *sounds*) nil))

(defun index-sound (name)
  (cond ((equal name 'click) 0)
        ((equal name 'whisk) 1)
        ((equal name 'snap)  2)
        ((equal name 'radar) 3)
        ((equal name 'chirp) 4)))

(defun load-sounds ()
  (setf *sounds* (list))
  (loop for x in (first-n-integers *number-of-sounds*) do
        (setf *sounds* (append *sounds* (list (#_GetResource "snd " (+ 128 x)))))))

;--------------- PEN OPS ------------------------------------

(defmethod pen-mode ((view simple-view))
  (elt *pen-modes* (pref (wptr view) windowRecord.pnmode)))

(defmethod set-pen-mode ((view simple-view) new-mode)
  (with-port (wptr view) (#_PenMode (mode-arg new-mode))))

(defmethod pen-pattern ((view simple-view) &optional
                        (save-pat (make-record (:pattern :storage :pointer))))
  (rlet ((state :PenState))
    (with-focused-view view
      (#_GetPenState state))
    (copy-record
     (pref state PenState.pnpat) (:pattern :storage :pointer) save-pat)))

(defmethod set-pen-pattern ((view simple-view) new-pattern)
  (with-port (wptr view)
    (#_PenPat new-pattern))
  new-pattern)

(defvar *32-bit-qd-pen-modes*
  '((:blend . 32)
    (:addPin . 33)
    (:addOver . 34)
    (:subPin . 35)
    (:transparent . 36)
    (:adMax . 37)
    (:subOver . 38)
    (:adMin . 39)
    (:hilite . 50)))

(defun mode-arg (thing)
  (or
   (and (fixnump thing) (<= 0 thing 64) thing)
   (position thing *pen-modes*)
   (cdr (assq thing *32-bit-qd-pen-modes*))
   (error "Unknown pen mode: ~a" thing)))

;--------------- COORDINATE OPS ------------------------------------

(defun LtGtL (thing view1 view2)
  (GtL (LtG thing view1) view2))

;--------------- LIST OPS ------------------------------------

(defun first-n-integers (n)
  (let ((x    0)
        (list nil))
    (loop while (< x n) do
          (setf list (append list (list x)))
          (setf x (+ x 1)))
    list))

(defun match? (target lineup)
  (let ((result nil))
    (loop for element in lineup do
          (if (equal element target)
            (setf result t)))
    result))

;--------------- MATH OPS ------------------------------------

(defun x-to-the-y (x y)
  (exp (* y (log x))))